/* bulkman.c */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>

#include <isode/quipu/ds_search.h>
#include <isode/quipu/modify.h>

#include "algoname.h"
#include "bulkloadglobal.h"
#include "bulkloadattrnames.h"
#include "configuration.h"
#include "filter.h"
#include "list.h"
#include "modify.h"
#include "namelist.h"
#include "readsearch.h"
#include "processandupdate.h"
#include "processbulkandupdatedir.h"
#include "subdirectory.h"
#include "util.h"

extern	struct	bulkLoadConfigStruct	bulkLoadConfig ;
extern	struct	managementConfigStruct	managementConfig ;

#define	MANAFILENAME	"bulkloadmanparseA"
#define	MANBFILENAME	"bulkloadmanparseB"

int
defineAndActionAnOOUandLSearch( outfp, parentDN, searchListpp )
FILE *	outfp ;
DN	parentDN ;
struct 	namelist ** searchListpp ;
{
int	retVal = SEARCHNOTOK ;
struct 	namelist * attrsToBeReturnedp = NULLLIST ;
struct	ds_search_arg	 sarg;
struct	ds_search_result sresult;
struct	DSError 	 serror ;
char *	errorcp ;

	if( parentDN == NULLDN )
	    return( retVal ) ;

	addToList( &attrsToBeReturnedp, BL_ORGANIZATION ) ;
	addToList( &attrsToBeReturnedp, BL_ORGANIZATIONAL_UNIT ) ;
	addToList( &attrsToBeReturnedp, BL_LOCALITY ) ;

	sarg = *fillMostSearchArgs( parentDN, SRA_WHOLESUBTREE, attrsToBeReturnedp, FALSE ) ;

	sarg.sra_filter = orfilter() ;
	sarg.sra_filter->FUFILT = eqfilter( FILTERITEM_EQUALITY, BL_OBJECT_CLASS, BL_ORG_OBJECT_CLASS ) ;
	sarg.sra_filter->FUFILT->flt_next = eqfilter( FILTERITEM_EQUALITY, BL_OBJECT_CLASS, BL_ORG_UNIT_OBJECT_CLASS ) ;

	retVal = RetrieveAndListAttrs( searchListpp, &sarg, &sresult, &serror ) ;

        if (retVal != SEARCHOK)
 	{
  	    logListSuccess( LIST_ERROR, "bulkManSourceReportDetails", 0) ;

	    errorcp = error2pstr( &serror ) ; /* ds_error_free() called within this procedure */

	    if( errorcp != NULLCP )
	    {
	    	fprintf( outfp, "#ERROR : %s\n", errorcp ) ;
		free( errorcp ) ;
	    }
 	}
	else
	{
	    logListSuccess( LIST_OK, "bulkManSourceReportDetails", listlen( *searchListpp )) ;
	}

	freeListpp( &attrsToBeReturnedp ) ;
	freeSearchArgs( &sarg ) ;
	
	return( retVal ) ; /* SEARCHNOTOK, SEARCHOK,... [bulkloadglobal.h] */
}


void
generateSourceReport( sourceReportfp )
FILE *	sourceReportfp ;
{
struct 	namelist * searchListp = NULLLIST, * listp ;
struct  dataSourceLevel * baseDataSourceLevelp = NULLSOURCELEVEL ;

	if( managementConfig.sourceReportSubTreeParentDN == NULLDN )
	{
	    fprintf( sourceReportfp, "#ERROR source report Sub-tree parent DN has been corrupted\n" ) ;
	    return ;
	}

	if( defineAndActionAnOOUandLSearch( sourceReportfp, managementConfig.sourceReportSubTreeParentDN, &searchListp ) == SEARCHOK )
	{
	    /* Now traverse any organisations or units that have been found.. */

	    for( listp = searchListp ;  listp != NULLLIST ; listp = listp -> next )
	    {
		(void)retrieveBulkDirectoryDetails( listp -> entryDN, &baseDataSourceLevelp ) ;

		if( baseDataSourceLevelp != NULLSOURCELEVEL )
		{
		    fprintf( sourceReportfp, "\n####################################\n%s\n\n", listp->name ) ;

		    /* Output the master/create and delete source details.. */
		    reportSourceDetails( sourceReportfp, baseDataSourceLevelp ) ;
		    
		    free_BulkDirectoryDetailspp( &baseDataSourceLevelp ) ;
		}
	    }

	    freeListpp( &searchListp ) ;
	}
	
	return ;
}


void
processSourceReport( fatalInterrupt )
int	fatalInterrupt ;
{
FILE *	sourceReportfp ;

	if( fatalInterrupt == FALSE )
	{
	    /* Input the command parameters.. */
	    processSourceReportUpdates() ;

	    if( managementConfig.returnToMenu == TRUE )
	    	return ;

	    if( (sourceReportfp = fopen( managementConfig.sourceReportFileNamecp, "w" )) == (FILE *)NULL)
	    	fprintf( stderr, "Attempt to open the report file failed\n" ) ;

	    else {

	    	fprintf( sourceReportfp, "\t\tSOURCE REPORT\n\t\t=============\n\n" );

		fprintf( sourceReportfp, "Highest directory level to search for dataSource entries = %d\n\n",
			bulkLoadConfig.highestDataSourceSearchLevel ) ;

		generateSourceReport( sourceReportfp ) ;
		
		fclose( sourceReportfp ) ;	    
	    }
	
	}
}

/*
 * Search for entries that have a UTC time of the specified attribute type
 * at or before the specified UTC time.
 */

int
defineAndActionUTCtimeSearch( logfp, parentDN, searchListpp, attrTypecp, UTCtimecp )
FILE *	logfp ;
DN	parentDN ;
struct 	namelist ** searchListpp ;
char *	attrTypecp, * UTCtimecp ;
{
int	retVal = SEARCHNOTOK ;
struct 	namelist * attrsToBeReturnedp = NULLLIST ;
struct	ds_search_arg	 sarg;
struct	ds_search_result sresult;
struct	DSError 	 serror ;
char *	errorcp ;

	if( parentDN == NULLDN )
	    return( retVal ) ;

	addToList( &attrsToBeReturnedp, attrTypecp ) ;

	sarg = *fillMostSearchArgs( parentDN, SRA_WHOLESUBTREE, attrsToBeReturnedp, FALSE ) ;

	sarg.sra_filter = orfilter() ;
	sarg.sra_filter->FUFILT = eqfilter( FILTERITEM_LESSOREQUAL, attrTypecp, UTCtimecp ) ;

	retVal = RetrieveAndListAttrs( searchListpp, &sarg, &sresult, &serror ) ;

        if (retVal != SEARCHOK)
 	{
  	    logListSuccess( LIST_ERROR, "bulkManUTCtimeEntries", 0) ;

	    errorcp = error2pstr( &serror ) ; /* ds_error_free() called within this procedure */

	    if( errorcp != NULLCP )
	    {
	    	fprintf( logfp, "#ERROR : %s\n", errorcp ) ;
		free( errorcp ) ;
	    }
 	}
	else
	{
	    logListSuccess( LIST_OK, "bulkManUTCtimeEntries", listlen( *searchListpp )) ;
	}

	freeSearchArgs( &sarg ) ;
	freeListpp( &attrsToBeReturnedp ) ;

	return( retVal ) ; /* SEARCHNOTOK, SEARCHOK,... [bulkloadglobal.h] */
}

int
retrieveExpiredEntries( deleteOutfp, deleteLogfp, parentDN, UTCtimecp )
FILE *	deleteOutfp, * deleteLogfp ;
DN	parentDN ;
char *	UTCtimecp ;
{
int	entriesForDeletion = FALSE ;
struct	namelist * entryListp, * searchListp = NULLLIST ;

	if( defineAndActionUTCtimeSearch( deleteLogfp, parentDN, &searchListp, BL_DELETE_ENTRY_AFTER, UTCtimecp ) == SEARCHOK )
	{	    
	    for( entryListp = searchListp ; entryListp != NULLLIST ; entryListp = entryListp -> next )
	    {
	    	if( entryListp -> name != NULLCP )
		{
		    fprintf( deleteOutfp, "%s\n", entryListp -> name ) ;
		    entriesForDeletion = TRUE ;
		}
	    }

	    freeListpp( &searchListp ) ;
	}
	
	return( entriesForDeletion ) ;
}


/*
 * Attempts to apply only a single rootedAt for the entries in each subtree level.
 */

void
outputEntryToFile( outputLogfp, entryListp, entryDN, firstCall )
FILE *	outputLogfp ;
struct	namelist * entryListp ;
DN	entryDN ;
int	firstCall ;
{
int	rootedAtReq = FALSE ;
char *	dncp, * rdncp, * avscp, * attrTcp ;
Attr_Sequence as ;
DN	entryDNcopy, currDNcomp, prevDNcomp ;
static	DN	bufferedParentEntryDN = NULLDN ;

	/* Dump a rootedAt statement and the RDN */
	if( (entryDNcopy = dn_cpy( entryDN )) != NULLDN )
	{
	    currDNcomp = entryDNcopy, prevDNcomp = NULLDN ;

	    /* Find the last two DN components */
	    while( currDNcomp != NULLDN )
	    {
		if( currDNcomp -> dn_parent != NULLDN )
		{
	    	    prevDNcomp = currDNcomp ;
		    currDNcomp = currDNcomp -> dn_parent ;
		}
		else break ;
	    }

	    if( currDNcomp != NULLDN && prevDNcomp != NULLDN )
	    {
	   	/* Seperate off the last DN component (pointed to by currDNcomp) */
	        prevDNcomp -> dn_parent = NULLDN ;

		if( firstCall == TRUE )
		{
		    if( bufferedParentEntryDN != NULLDN )
			dn_free( bufferedParentEntryDN ) ; 
		    
		    bufferedParentEntryDN = dn_cpy( entryDNcopy ) ;

		    rootedAtReq = TRUE ;
		}
	
		if( bufferedParentEntryDN != NULLDN && dn_cmp( bufferedParentEntryDN, entryDNcopy ) != 0 )
		{
		    dn_free( bufferedParentEntryDN ) ; bufferedParentEntryDN = dn_cpy( entryDNcopy ) ;
		    
		    rootedAtReq = TRUE ;
		}
		
		/* The rootedAt DN is obtained from the remaining DN components.. */
		if( rootedAtReq == TRUE && (dncp = dn2pstr( entryDNcopy )) != NULLCP )
		{
		    fprintf( outputLogfp, "\n\nrootedAt= %s\n", dncp ) ;
		    free( dncp ) ;
		}
	    }

	    /* Now output the RDN.. */
	    if( currDNcomp != NULLDN && currDNcomp -> dn_rdn != NULLRDN &&
		(rdncp = rdn2pstr( currDNcomp -> dn_rdn ))   != NULLCP    )
	    {
		fprintf( outputLogfp, "\n%s\n", rdncp ) ;
		free( rdncp ) ;
	    }

	    if( currDNcomp != NULLDN && currDNcomp != entryDNcopy )
		dn_free( currDNcomp ) ;

	    dn_free( entryDNcopy ) ;
	}

	/* Now output the attributes (if any).. */
	for( as = entryListp -> ats ; as != NULLATTR ; as = as -> attr_link )
	{
	    if( (as -> attr_type != NULLAttrT) && (attrTcp = AttrT2pstr( as -> attr_type )) != NULLCP )
	    {
		if( (as->attr_value != NULLAV) && (avscp = avs2pstr( as->attr_value )) != NULLCP )
		{
		    fprintf( outputLogfp, "%s = %s", attrTcp, avscp ) ;
		    free( avscp ) ;
		}

		free( attrTcp ) ;
	    }
	}
	    
	return ;
}


void
deleteExpiredEntries( deleteEntriesLogfp, deleteEntryDN, deleteEntrycp, firstCallFromThisOption )
FILE *	deleteEntriesLogfp ;
DN	deleteEntryDN ;
char *	deleteEntrycp ;
int	firstCallFromThisOption ;
{
char *	errorMsgcp ;
struct	namelist * readEntryListp = NULLLIST ;

	if( readEntry( deleteEntryDN, &readEntryListp, NULLLIST, TRUE ) == OK &&
	    readEntryListp != NULLLIST )
	{
	    resetprint( "About to attempt to remove the entry..\n" ) ;
	    pageprint(  "%s\n", deleteEntrycp ) ;
	
	    if( RemoveEntry( deleteEntryDN, &errorMsgcp ) != OK )
	    {
	    	if( errorMsgcp != NULLCP )
	    	{
	    	    fprintf( deleteEntriesLogfp, "# Could not DELETE the following entry\n" ) ;
	    	    fprintf( deleteEntriesLogfp, "# %s\n", deleteEntrycp ) ;
	    	    fprintf( deleteEntriesLogfp, "# %s...\n", errorMsgcp ) ;
	    	    free( errorMsgcp ) ;
		}
	    }
	    else
	    {
	    	/* Dump the previous contents to the log file now that the entry has been deleted.. */
		outputEntryToFile( deleteEntriesLogfp, readEntryListp, deleteEntryDN, firstCallFromThisOption ) ;
		fflush( deleteEntriesLogfp ) ;
	    }

	    freeListpp( &readEntryListp ) ;
	}
	else 
	{
	    fprintf( deleteEntriesLogfp, "#ERROR Could not READ the following entry's contents before a delete could be attempted\n" ) ;
	    fprintf( deleteEntriesLogfp, "# %s\n", deleteEntrycp ) ;
	}
		
	return ;
}

void
processExpiredEntries( fatalInterrupt )
int	fatalInterrupt ;
{
FILE *	expiredEntriesAfp, * expiredEntriesBfp, * expiredEntriesfp, * deleteEntriesLogfp, *infp, *outfp ;
int	fromAtoB = TRUE, deleteEntriesParsed = FALSE, entriesToBeDeleted=FALSE, errorOccurred=FALSE ;
int	firstIteration, numOfEntriesToBeDeleted, firstCall ;
char	deleteEntryLine[ SMALLBUFLEN ], * linecp ;
DN	deleteDN ;
char *	manAfilecp, * manBfilecp, manAfilebuff[ SMALLBUFLEN ], manBfilebuff[ SMALLBUFLEN ] ;

	sprintf( manAfilebuff, "%sXXXXXX", MANAFILENAME ) ;
	manAfilecp = mktemp( manAfilebuff ) ;

	sprintf( manBfilebuff, "%sXXXXXX", MANBFILENAME ) ;
	manBfilecp = mktemp( manBfilebuff ) ;

	if( fatalInterrupt == FALSE )
	{
	    /* Input the command parameters.. */
	    processExpiredEntriesUpdates() ;

	    if( managementConfig.returnToMenu == TRUE )
	    	return ;

	    if( (deleteEntriesLogfp = fopen( managementConfig.expiredEntriesLogFileNamecp, "w" )) == (FILE *)NULL)
	    	fprintf( stderr, "Attempt to open the log file failed\n" ) ;

	    else if( managementConfig.expiredEntriesUTCtimecp == NULLCP )
	    	fprintf( stderr, "The delete entry after time was missing\n" ) ;

	    else {

	    	fprintf( deleteEntriesLogfp, "#\t\tDELETE ENTRIES LOG FILE\n#\t\t=======================\n\n" );

		if( (expiredEntriesAfp = fopen( manAfilecp, "w" )) == (FILE *)NULL )
		    fprintf( stderr, "Attempt to open the parse file failed\n" ) ;

		else {

		    entriesToBeDeleted = retrieveExpiredEntries( expiredEntriesAfp, deleteEntriesLogfp,
	 			    		managementConfig.expiredEntriesSubTreeParentDN, 
 						managementConfig.expiredEntriesUTCtimecp ) ;

		    fclose( expiredEntriesAfp ) ;

		    deleteEntriesParsed = FALSE ; firstIteration = TRUE ;

		    while( entriesToBeDeleted == TRUE && deleteEntriesParsed == FALSE )
		    {
		    	if( (expiredEntriesAfp = fopen( manAfilecp, (fromAtoB == TRUE) ? "r" : "w" )) == (FILE *)NULL ||
		            (expiredEntriesBfp = fopen( manBfilecp, (fromAtoB == TRUE) ? "w" : "r" )) == (FILE *)NULL   )
			{
			    fprintf( stderr, "Attempt to open the parse file failed\n" ) ;

			    if( expiredEntriesAfp != ((FILE *)NULL) )
			    	fclose( expiredEntriesAfp ) ;
				
		    	    break ;
			}
			else
			{
			    infp  = (fromAtoB == TRUE) ? expiredEntriesAfp : expiredEntriesBfp ;
			    outfp = (fromAtoB == TRUE) ? expiredEntriesBfp : expiredEntriesAfp ;
			    
			    fprintf( outfp, "#\n# Please review this list of entries that are to be deleted\n#\n" ) ;
			    fprintf( outfp, "#\n# Remove those that MUST NOT be deleted\n#\n" ) ;
			    fprintf( outfp, "# On an error condition search for \"#ERROR\"\n#\n" ) ;

			    errorOccurred = FALSE ; entriesToBeDeleted = FALSE ;

			    numOfEntriesToBeDeleted = 0 ;
			
			    /* Attempt to parse the list of entries to be deleted.. */
			    while( fgets( deleteEntryLine, SMALLBUFLEN, infp ) != NULL )
			    {
				if( strncasecmp( deleteEntryLine, "#ERROR", 6 ) != 0 )
			    	    fprintf( outfp, "%s", deleteEntryLine ) ;

				if( *deleteEntryLine != '#' && !isspace( *deleteEntryLine ) )
				{
				    linecp = TidyString( deleteEntryLine ) ;
				    
				    if( (linecp == NULLCP) || (deleteDN = str2dn( linecp )) == NULLDN )
				    {
				    	fprintf( outfp, "#ERROR\n#ERROR *** %s\n#ERROR\n", "Invalid distinguished name" ) ;
				    	errorOccurred = TRUE ;
				    }
				    else 
				    {	
				    	entriesToBeDeleted = TRUE ;
					numOfEntriesToBeDeleted++ ;
					dn_free( deleteDN ) ;
				    }
				}
			    }

			    fclose( infp ) ; fclose( outfp ) ;

			    if( firstIteration == FALSE && errorOccurred == FALSE )
			    {
			    	resetprint( "" ) ; pageprint( "[ %d ] entries to be deleted\n", numOfEntriesToBeDeleted ) ;
			    	deleteEntriesParsed = TRUE ;
			    }
			    else
			    {	
			        /* Display "expired" entry DNs upon error or during the first iteration
				   (to allow for confirmation by the "enterYesNo" statment that follows) */
				   
			    	editFile( (fromAtoB == TRUE) ? manBfilecp : manAfilecp ) ;

				/* Prepare for the next parsing iteration.. */
			    	fromAtoB = (fromAtoB == TRUE) ? FALSE : TRUE ;
			    }
			}

			firstIteration = FALSE ;
		    }

		    if( entriesToBeDeleted == TRUE && deleteEntriesParsed == TRUE )
		    {
		    	if( enterYesNo( "Are you sure that these entries should be deleted ?", "(y/n)" ) == 'y' )
		    	{
			    if( (expiredEntriesfp = fopen( (fromAtoB==TRUE) ? manBfilecp : manAfilecp, "r" )) == (FILE *)NULL )
			    {
			    	fprintf( stderr, "Attempt to open the parse file failed\n" ) ;
			    }
			    else 
			    {
			    	firstCall = TRUE ;

			    	fprintf( deleteEntriesLogfp, "\nsource=bulk\n\n" );
				
		    	    	/* Run through any remaining entries, deleting the corresponding entries.. */
				while( fgets( deleteEntryLine, SMALLBUFLEN, expiredEntriesfp ) != NULL )
			    	{
				    if( strncasecmp( deleteEntryLine, "#ERROR", 6 ) != 0 &&
					*deleteEntryLine != '#' && !isspace( *deleteEntryLine ) )
				    {
				    	linecp = TidyString( deleteEntryLine ) ;
				    
				    	if( (linecp != NULLCP) && (deleteDN = str2dn( linecp )) != NULLDN )
				        {
					    deleteExpiredEntries( deleteEntriesLogfp, deleteDN, linecp, firstCall ) ;
					    dn_free( deleteDN ) ;
					    firstCall = FALSE ;
					}
				    }
				}

			    	fclose( expiredEntriesfp ) ;
			    }
		    	}
		    }
		    else fprintf( deleteEntriesLogfp, "# There were no remaining expired entries\n" ) ;
		    
    		    unlink( manAfilecp ) ; unlink( manBfilecp ) ;
		}

		fclose( deleteEntriesLogfp ) ;
	    }
	}

	return ;
}

/*
 * Note: "setTheValidationCheckAtForLeafNode()", "dumpSubtreeNode()", "bulkManAddAttributeOrValue()" and
 * "exchangeMasterUpAtforNonMasterUpAt()"
 * have common parameters since the functions are passed to "applyOperationToSubTreeNodes()" to be called.
 */

int
dumpSubtreeNode( outfp, nodeDN, dummycp, dummyAttrT, dummyAttrV, firstCallFromThisOption )
FILE *	outfp   ;
DN	nodeDN  ;
char *	dummycp ;
AttributeType	dummyAttrT ;
AttributeValue	dummyAttrV ;
int	firstCallFromThisOption ;
{
int	retVal = OK ;
struct	namelist * readEntryListp = NULLLIST ;
char *	DNcp ;

	if( readEntry( nodeDN, &readEntryListp, NULLLIST, TRUE ) == OK &&
	    readEntryListp != NULLLIST )
	{
	    /* Dump the previous contents to the log file now that the entry has been deleted.. */
	    outputEntryToFile( outfp, readEntryListp, nodeDN, firstCallFromThisOption ) ;
	
	    freeListpp( &readEntryListp ) ;	    
	}
	else 
	{
	    DNcp = dn2pstr( nodeDN ) ;	    
	
	    fprintf( stderr, "#ERROR Could not READ the following entry's contents\n" ) ;
	    fprintf( stderr, "# %s\n", (DNcp != NULLCP) ? DNcp : "INVALID ENTRY DN" ) ;
	  
	    fprintf( outfp, "#ERROR Could not READ the following entry's contents\n" ) ;
	    fprintf( outfp, "# %s\n", (DNcp != NULLCP) ? DNcp : "INVALID ENTRY DN" ) ;

	    if( DNcp != NULLCP )
	    	free( DNcp ) ;

	    /* retval = NOTOK ; treat this a warning condition for the time being */
	}
	
	
	return( retVal ) ;
}


void
processDumpSubtree( fatalInterrupt )
int	fatalInterrupt ;
{
FILE *	dumpSubtreefp ;

	if( fatalInterrupt == FALSE )
	{
	    /* Input the command parameters.. */
	    processDumpSubtreeUpdates() ;

	    if( managementConfig.returnToMenu == TRUE )
	    	return ;

	    if( (dumpSubtreefp = fopen( managementConfig.dumpSubtreeFileNamecp, "w" )) == (FILE *)NULL)
	    	fprintf( stderr, "Attempt to open the output file failed\n" ) ;
	    else
	    {
		if( applyOperationToSubTreeNodes( dumpSubtreefp, managementConfig.dumpSubtreeSubTreeParentDN,
			      dumpSubtreeNode, FALSE, NULLCP, NULLAttrT, NULLAttrV ) == NOTOK )
		    fprintf( stderr, "Dump subtree failure for\n%s\n", managementConfig.dumpSubtreeSubTreeParentDNcp ) ;
					    

		fclose( dumpSubtreefp ) ;
	    }
	}

	return ;
}
	
/*
 * Note: "setTheValidationCheckAtForLeafNode()", "dumpSubtreeNode()", "bulkManAddAttributeOrValue()" and
 * "exchangeMasterUpAtforNonMasterUpAt()"
 * have common parameters since the functions are passed to "applyOperationToSubTreeNodes()" to be called.
 */

int
bulkManAddAttributeOrValue( outfp, nodeDN, dummycp, at, av, firstCallFromThisOption )
FILE *	outfp ;
DN	nodeDN ;
char *	dummycp ;
AttributeType	at ;
AttributeValue	av ;
int	firstCallFromThisOption ;
{
	addAttrOrValue( nodeDN, at, av, TRUE ) ;

	return( OK ) ;
}


void
processAddValueToSubTreeNodes( fatalInterrupt )
int	fatalInterrupt ;
{
	if( fatalInterrupt == FALSE )
	{
	    /* Input the command parameters.. */
	    processAddValueToSubTreeUpdates() ;

	    if( managementConfig.returnToMenu == TRUE )
	    	return ;

	    if( applyOperationToSubTreeNodes( ((FILE *)NULL), managementConfig.addValueToSubTreeParentDN,
			bulkManAddAttributeOrValue, managementConfig.addValueToJustLeafNodes, NULLCP,
 			managementConfig.addValueToSubTreeAt, managementConfig.addValueToSubTreeAv ) == NOTOK )
		    fprintf( stderr, "Add attribute failure for\n%s\n", managementConfig.addValueToSubTreeParentDNcp ) ;
	}

	return ;
}
	


/*
 * Note: "setTheValidationCheckAtForLeafNode()", "dumpSubtreeNode()", "bulkManAddAttributeOrValue()" and
 * "exchangeMasterUpAtforNonMasterUpAt()"
 * have common parameters since the functions are passed to "applyOperationToSubTreeNodes()" to be called.
 */

int
setTheValidationCheckAtForLeafNode( erroroutfp, nodeDN, UTCtimecp, dummyAttrT, dummyAttrV, firstCallFromThisOption )
FILE *	erroroutfp  ; /* May be NULL */
DN	nodeDN ;
char *	UTCtimecp ;
AttributeType	dummyAttrT ;
AttributeValue	dummyAttrV ;
int	firstCallFromThisOption ;
{
int	retVal = OK ;
char *	errorMsgcp, * DNcp ;
RDN	finalNodeRDN ;
struct	DSError	 * dserrorp ;
static	struct  	entrymod * baseModNodep = NULLMOD, * currModNodep ;
static	AttributeType	validationCheckAtat = NULLAttrT, cnat = NULLAttrT, uidat = NULLAttrT ;
static	AttributeValue	validationCheckAtav = NULLAttrV ;
static	char * lastUTCtimecp = NULLCP ;

	if( lastUTCtimecp == NULLCP )
	{
	    lastUTCtimecp =  strdup( UTCtimecp ) ;	
	    cnat =  str2AttrT( BL_COMMON_NAME ) ;
	    uidat = str2AttrT( BL_UID ) ;

	    if( (validationCheckAtat = str2AttrT( BL_VALIDATION_CHECK_AT )) != NULLAttrT )
	    	validationCheckAtav = str2AttrV( UTCtimecp, validationCheckAtat->oa_syntax ) ;
	}
	else
	{
	    /* Has the UTC time changed.. */
	    if( strcmp( lastUTCtimecp, UTCtimecp ) != 0 )
	    {
	    	/* Buffer the new UTC time.. */
	    	free( lastUTCtimecp ) ;
		lastUTCtimecp = strdup( UTCtimecp ) ;
	    
	    	if( validationCheckAtav != NULLAttrV )
		{
		    AttrV_free( validationCheckAtav ) ; 
		    validationCheckAtav = str2AttrV( UTCtimecp, validationCheckAtat->oa_syntax ) ;
		}

	    	if( baseModNodep != NULLMOD )
		{
		    free_attr_mod_list( baseModNodep ) ; baseModNodep = NULLMOD ;
		}
	    }
	}

	if( cnat == NULLAttrT || uidat == NULLAttrT || validationCheckAtat == NULLAttrT || validationCheckAtav == NULLAttrV )
 	    return( NOTOK ) ;

	if( baseModNodep == NULLMOD )
	{
	    /* Form the modification request that is to be applied to each leaf node.. */
	    currModNodep = NULLMOD ;

	    currModNodep = create_attr_mod( EM_REMOVEATTRIBUTE, validationCheckAtat, NULLAttrV ) ;
	    append_attr_mod( &baseModNodep, currModNodep ) ;
	
	    currModNodep = create_attr_mod( EM_ADDATTRIBUTE, validationCheckAtat, validationCheckAtav ) ;
	    append_attr_mod( &baseModNodep, currModNodep ) ;
	}

	if( (finalNodeRDN = findLastRDN( nodeDN )) != NULLRDN )
	{
	    /* Check that the leaf node should have the validationChechAt attribute added.. */
	    if( findAttrTypeRdnComp( cnat, finalNodeRDN ) != NULLRDN ||
	        findAttrTypeRdnComp( uidat,finalNodeRDN ) != NULLRDN   )
	    {
		if((DNcp = dn2pstr( nodeDN )) != NULLCP )
		{
		    resetprint("Modifying validationCheckAt for\n%s\n", DNcp ) ;
		}
		
		/* try removing the validationCheckAt attribute before adding the new value.. */
		if( modifyDirectory( baseModNodep, nodeDN, &dserrorp ) != OK )
		{
		    ds_error_free( dserrorp ) ;

		    /* Try an add attribute request (will fail if the attribute is already present.. */
		    if( modifyDirectory( currModNodep, nodeDN, &dserrorp ) != OK )
		    {
		   	if( (errorMsgcp = error2pstr( dserrorp )) != NULLCP )
		    	{
			    pageprint( "Could not update the \"validationCheckAt\" for\n%s\n", (DNcp != NULLCP) ? DNcp : "" );

			    if( erroroutfp != ((FILE *)NULL) )
			    {
			    	fprintf( erroroutfp, "\n#ERROR %s\n", (DNcp != NULLCP) ? DNcp : "" ) ;
				fprintf( erroroutfp, "# %s\n", errorMsgcp ) ;
			    }

			    free( errorMsgcp ) ;
			}
		    }
		}

		if( DNcp != NULLCP )
		    free( DNcp ) ;		
	    }
	}

	return( retVal ) ;
}


void
processResetValidCheckAt( fatalInterrupt )
int	fatalInterrupt ;
{
FILE *	resetValidChkAtLogfp ;

	if( fatalInterrupt == FALSE )
	{
	    /* Input the command parameters.. */
	    processResetValidCheckAtUpdates() ;

	    if( managementConfig.returnToMenu == TRUE )
	    	return ;

	    if( (resetValidChkAtLogfp = fopen( managementConfig.resetValidChkAtLogFileNamecp, "w" )) == (FILE *)NULL )
	    	fprintf( stderr, "Attempt to open the output file failed\n" ) ;
	    else
	    {
	    	resetprint( "validationCheckAt UTCtime used = %s\n", managementConfig.resetValidChkAtUTCtimecp );

		fprintf( resetValidChkAtLogfp, "# validationCheckAt UTCtime used = %s\n", managementConfig.resetValidChkAtUTCtimecp );

	    	if( applyOperationToSubTreeNodes( resetValidChkAtLogfp, managementConfig.resetValidChkAtSubTreeParentDN,
						      setTheValidationCheckAtForLeafNode, TRUE,
 						      managementConfig.resetValidChkAtUTCtimecp, NULLAttrT, NULLAttrV ) == NOTOK )
	    	{
		    fprintf( stderr, "The reseting of validation check at failed for\n%s\n", managementConfig.resetValidChkAtSubTreeParentDNcp ) ;
		}

		fclose( resetValidChkAtLogfp ) ;
	    }
	}

	return ;
}


/*
 * This is strictly a bulk management procedure.
 *
 * Search for the algorithmic name, if found compare the
 * source file entry's attributes against the Directory entry's.
 * Also update the Directory entry's "validationCheckAt" attribute.
 */

int
attemptMatchAgainstDirectory( baseRAApp, OFD, sourceNamecp )
struct  rootedAtAttributes ** baseRAApp ;
struct  outputFileDetails OFD[] ;
char *	sourceNamecp ;
{
int	retVal = OK, matchFound = FALSE ;
char *	DNcp, * DNsubDirListcp, * UTCtimecp ;
struct 	namelist * searchListp = NULLLIST ;
char	diagnostics[ ERRORBUFSIZE ] ;
Attr_Sequence	as = NULLATTR ;
AV_Sequence	avs = NULLAV ;
struct  timeval  t,  *tp=&t ;
struct  timezone tz, *tzp=&tz ;
struct  tm * tmp ;
UTCtime utc, * utcp = &utc ;
struct  rootedAtAttributes * currRAAp = *baseRAApp ;
struct	attributeLine * currALp ;
static	int	subDirListInitialised = FALSE ;
static	struct	subDirListDetails rootedAtSubDirListDetails ;
static	DN prevDN = NULLDN ;

	diagnostics[ 0 ] = '\0' ;

	if( prevDN == NULLDN || dn_cmp( prevDN, currRAAp->rootedAtDN ) != 0 )
	{
	    DNcp = dn2pstr( currRAAp -> rootedAtDN ) ;

	    setUpOutFilesRA( OFD, (DNcp != NULLCP) ? DNcp : "INVALID ROOTEDAT DN" ) ;

	    if( DNcp != NULLCP )
		free( DNcp ) ;

	    if( prevDN != NULLDN )
	    	dn_free( prevDN ) ;

	    prevDN = dn_cpy( currRAAp->rootedAtDN ) ;
	}
	
	setUpOutFilesRDN( OFD, currRAAp -> rdnLinecp ) ;

	resetprint( "BULKMAN is currently processing %s\n", currRAAp -> rdnLinecp );

	if( subDirListInitialised == FALSE )
	{
	    initSubDirectoryList( &rootedAtSubDirListDetails ) ;
	    subDirListInitialised = TRUE ;
	}

	if( checkForAndExpandSubDirList( &rootedAtSubDirListDetails, currRAAp -> rootedAtDN, diagnostics ) == NOTOK )
	{
	    pageprint( diagnostics ) ;
	    transferRootedAtRDNattrs( OFD, OUTPUTPROBLEMFILE, currRAAp, diagnostics, NULLCP ) ;

	    return( NOTOK ) ;
	}

	/* Initiate an algorithmic name search (all of a matched entries attributes are to be returned).. */

	if( defineAndActionAlgoNameAttrSearch( sourceNamecp, currRAAp->rootedAtDN, currRAAp->rdn, 
			NULLLIST, TRUE, &searchListp, &rootedAtSubDirListDetails, diagnostics ) == SEARCHOK )
	{
	    if( searchListp == NULLLIST )
	    {
	    	/* An entry with the required algorithmic name could not be found.. */
		strcpy( diagnostics, "Search failed to find a directory entry with an algorithmic name for this entry" ) ;
		transferRootedAtRDNattrs( OFD, OUTPUTPROBLEMFILE, currRAAp, diagnostics, NULLCP ) ;
	    }
	    else
	    {
		if( rootedAtSubDirListDetails.subDirListUsed != SUBDIRLISTNOTUSED &&
	    	    rootedAtSubDirListDetails.rootedAtDN     != NULLDN )
		{
		    if( (DNsubDirListcp = extractParentDNandReturnDNcp( searchListp -> entryDN )) != NULLCP )
		    {
	    	    	setUpOutFilesSubDirListResolvedDN( OFD, DNsubDirListcp ) ;

		    	free( DNsubDirListcp ) ;
		    }
		}

		/* An entry has been found ! */

		if( !gettimeofday( tp, tzp) )
		{
		    tmp = localtime( &((time_t)(tp -> tv_sec)) ) ;
		    tm2ut( tmp, utcp ) ;
	            UTCtimecp = time2str( utcp, 0 ) ;

		    /* Update the entries validationCheckAt attribute */
		    (void)setTheValidationCheckAtForLeafNode( OFD[OUTPUTPROBLEMFILE].outFilep, 
		    					searchListp->entryDN, UTCtimecp, NULLAttrT, NULLAttrV, FALSE ) ;
            	}

		/* Run through the directory entry's attributes checking for the source entry's attributes.. */
		for( currALp = currRAAp -> attrLinep ; currALp != NULLALP ; currALp = currALp -> nextAttrLinep )
		{
		    /* The operation is not applied to ACL attributes.. */
		    if( strcmp( attr2name( currALp->at, OIDPART ), BL_ACCESS_CONTROL_LIST ) != 0 )
		    {
			matchFound = FALSE ;
			diagnostics[ 0 ] = '\0' ;

			for( as = searchListp->ats ; as != NULLATTR ; as = as -> attr_link )
			{
			    if( AttrT_cmp( as -> attr_type, currALp -> at ) == 0 )
			    {
			    	for( avs = as -> attr_value ; avs != NULLAV ; avs = avs -> avseq_next )
				{
    				    if( AttrV_cmp( &(avs->avseq_av), currALp -> av ) == 0 )
					matchFound = TRUE ;
				}
			    }
			}

			if( matchFound == TRUE )
			{
			    transferAttribute( OFD, OUTPUTDONEFILE, currALp->attrLinecp, NULLCP, NULLCP ) ;
			}
			else
			{
			    strcpy( diagnostics, "Attribute value was not found in directory entry" ) ;
			    transferAttribute( OFD, OUTPUTPROBLEMFILE, currALp->attrLinecp, diagnostics,
				    (rootedAtSubDirListDetails.subDirListUsed != SUBDIRLISTNOTUSED) ? searchListp->name : NULLCP ) ;
			}
		    }
		}
		
		freeListpp( &searchListp ) ;
	    }
	}
	else transferRootedAtRDNattrs( OFD, OUTPUTPROBLEMFILE, currRAAp, diagnostics, NULLCP ) ;

	return( retVal ) ;
}


/* 
 * Run through the entries contained in the source file and attempt to
 * match them off against the algorithmic name attributes that are
 * already held within directory entries.
 */

void
processSourceValidation( fatalInterrupt )
int	fatalInterrupt ;
{
	if( fatalInterrupt == FALSE )
	{
	    /* Input the command parameters.. */
	    processSourceValidationUpdates() ;

	    if( managementConfig.returnToMenu == TRUE )
	    	return ;

	    /* Utilise the bulkload file processing to call the management operation "attemptMatchAgainstDirectory()" */
	    processInputFile( INTERACT_OPER_MODE, managementConfig.sourceValidationInputFileNamecp, TRUE ) ;
	}

	return ;
}

/*
 * Find any entries that have not be matched against during a source validation check.
 */

void
processCheckValidCheckAt( fatalInterrupt )
int	fatalInterrupt ;
{
FILE *	checkValidChkAtfp ;
struct	namelist * searchListp = NULLLIST, * entryListp ;
int	firstCall = TRUE ;

	if( fatalInterrupt == FALSE )
	{
	    /* Input the command parameters.. */
	    processCheckValidCheckAtUpdates() ;

	    if( managementConfig.returnToMenu == TRUE )
	    	return ;

	    if( (checkValidChkAtfp = fopen( managementConfig.checkValidChkAtLogFileNamecp, "w" )) == (FILE *)NULL)
	    	fprintf( stderr, "Attempt to open the output file failed\n" ) ;
	    else
	    {
		fprintf( checkValidChkAtfp, "#\t\tREPORT OF THE ENTRIES THAT HAVE NOT BEEN VALIDATED AGAINST\n#\t\t\
==========================================================\n\n" );

		if( defineAndActionUTCtimeSearch( checkValidChkAtfp, managementConfig.checkValidChkAtSubTreeParentDN, 
				&searchListp, BL_VALIDATION_CHECK_AT,
					managementConfig.checkValidChkAtUTCtimecp ) == SEARCHOK )
		{
		    for( entryListp = searchListp ; entryListp != NULLLIST ; entryListp = entryListp -> next )
	    	    {
	   	 	if( entryListp -> name != NULLCP )
			{
			    outputEntryToFile( checkValidChkAtfp, entryListp, entryListp->entryDN, firstCall ) ;
			    firstCall = FALSE ;
			}
		    }
		    
		    freeListpp( &searchListp ) ;
		}

		fclose( checkValidChkAtfp ) ;
	    }
	}

	return ;
}



/*
 * Note: "setTheValidationCheckAtForLeafNode()", "dumpSubtreeNode()", "bulkManAddAttributeOrValue()" and
 * "exchangeMasterUpAtforNonMasterUpAt()"
 * have common parameters since the functions are passed to "applyOperationToSubTreeNodes()" to be called.
 */

int
exchangeMasterUpAtforNonMasterUpAt( outfp, nodeDN, modecp, dummyat, dummyav, firstCallFromThisOption )
FILE *	outfp ;
DN	nodeDN ;
char *	modecp ;
AttributeType	dummyat ;
AttributeValue	dummyav ;
int	firstCallFromThisOption ;
{
struct	namelist * readMasterUpAtEntryListp = NULLLIST, * readAllEntryListp = NULLLIST ;
RDN	lastEntryRDN ;
Attr_Sequence	as  ;
AV_Sequence 	curravs, masterUpavs = NULLAV, nonMasterUpavs = NULLAV ;
AttributeType	at = NULLAttrT, * masterUpAtValueAttrTlist ;
AttributeValue	av = NULLAttrV ;
struct  entrymod * baseModNodep = NULLMOD, * currModNodep = NULLMOD ;
struct	DSError	 * dserrorp ;
char *	DNcp, * avcp, * errorMsgcp ;
int	nonMasterUpAtInDirectory, nonMasterUpToBeUpdated, i, attrTypeRejected, numMasterUpAtValues ;

static	struct	namelist * attrsToBeReturnedp = NULLLIST ;
static	AttributeType masterUpAtat = NULLAttrT, nonMasterUpAtat = NULLAttrT ;
static	struct	entrymod * baseRemoveMasterUpAtModNodep = NULLMOD ;
static	rejectAttrsTableInitialised = FALSE ;
#define NUMOFREJECTATTRT	11
static	AttributeType rejectAttrsTable[ NUMOFREJECTATTRT ] ;

	if( attrsToBeReturnedp == NULLLIST )
	{
	    addToList( &attrsToBeReturnedp, BL_MASTER_UPDATED_ATTR_TYPE ) ;
	    addToList( &attrsToBeReturnedp, BL_NON_MASTER_UPDATED_ATTR_TYPE ) ;
	}

	if( masterUpAtat == NULLAttrT )
	    masterUpAtat =  str2AttrT( BL_MASTER_UPDATED_ATTR_TYPE ) ;

	if( nonMasterUpAtat == NULLAttrT )
	    nonMasterUpAtat =  str2AttrT( BL_NON_MASTER_UPDATED_ATTR_TYPE ) ;


	if( baseRemoveMasterUpAtModNodep == NULLMOD )
	{
	    currModNodep = create_attr_mod( EM_REMOVEATTRIBUTE, masterUpAtat, NULLAttrV ) ;
	    append_attr_mod( &baseRemoveMasterUpAtModNodep, currModNodep ) ;
	}
	
	if( rejectAttrsTableInitialised == FALSE )
	{
	    rejectAttrsTable[ 0 ] = str2AttrT( "lastModifiedTime" ) ;
	    rejectAttrsTable[ 1 ] = str2AttrT( "lastModifiedBy" ) ;
	    rejectAttrsTable[ 2 ] = str2AttrT( BL_ACCESS_CONTROL_LIST ) ;
	    rejectAttrsTable[ 3 ] = str2AttrT( BL_MASTER_UPDATED_ATTR_TYPE ) ;
	    rejectAttrsTable[ 4 ] = str2AttrT( BL_NON_MASTER_UPDATED_ATTR_TYPE ) ;
	    rejectAttrsTable[ 5 ] = str2AttrT( BL_CREATE_SOURCE ) ;
	    rejectAttrsTable[ 6 ] = str2AttrT( BL_DELETE_SOURCE ) ;
	    rejectAttrsTable[ 7 ] = str2AttrT( BL_MASTERS ) ;
	    rejectAttrsTable[ 8 ] = str2AttrT( BL_ALGO_NAME ) ;
	    rejectAttrsTable[ 9 ] = str2AttrT( BL_DELETE_ENTRY_AFTER ) ;
	    rejectAttrsTable[ NUMOFREJECTATTRT - 1 ] = str2AttrT( BL_VALIDATION_CHECK_AT ) ;

	    rejectAttrsTableInitialised = TRUE ;
	}
	
	DNcp = dn2pstr( nodeDN ) ;

	/* retrieve the "masterUpdatedAttrType" and also the "nonMasterUpdatedAttrType" (if present) */
	if( readEntry( nodeDN, &readMasterUpAtEntryListp, attrsToBeReturnedp, FALSE ) == OK &&
	    readMasterUpAtEntryListp != NULLLIST )
	{
	    nonMasterUpAtInDirectory = FALSE ;
	    
	    /* check for the "masterUpdatedAttrType" attribute */
	    for( as = readMasterUpAtEntryListp->ats ; as != NULLATTR && (masterUpavs == NULLAV || nonMasterUpavs == NULLAV ) ;
	    	 as = as -> attr_link )
	    {
		if( AttrT_cmp( as -> attr_type, masterUpAtat ) == 0 )
		{
		    masterUpavs = as -> attr_value ;
		}
		else if( AttrT_cmp( as -> attr_type, nonMasterUpAtat ) == 0 )
		{
		    nonMasterUpAtInDirectory = TRUE ;
		    nonMasterUpavs = as -> attr_value ;
		    as -> attr_value = NULLAV ; /* To prevent it being freed by the results free */
		}
	    }

	    if( masterUpavs != NULLAV && (numMasterUpAtValues = countAVcomps( masterUpavs )) != 0 )
	    {
		resetprint( "BULKMAN is currently processing \n%s\n", DNcp );

		fprintf( outfp, "BULKMAN processed \"masterUpdateAttrType\" within\n%s\n", DNcp ) ;

		if( *modecp == 'r' )
		{
		    /* Remove the "masterUpdatedAttrType" attribute */
		    if( modifyDirectory( baseRemoveMasterUpAtModNodep, nodeDN, &dserrorp ) != OK )
		    {
		    	if( (errorMsgcp = error2pstr( dserrorp )) != NULLCP )
		    	{
		    	    fprintf( outfp, "%s\n", errorMsgcp ) ;
		    	    free( errorMsgcp ) ;
		    	}
		    }
		}
		else if( *modecp == 'e' )
		{
		    nonMasterUpToBeUpdated = FALSE ;
		
		    if( readEntry( nodeDN, &readAllEntryListp, NULLLIST, TRUE ) == OK &&
			readAllEntryListp != NULLLIST )
		    {
			baseModNodep = NULLMOD ;

			/* Hold the current "masterUpdatedAttrType" values as attribute types.. */
			if( (masterUpAtValueAttrTlist = (AttributeType *)malloc( (unsigned)(numMasterUpAtValues * sizeof(AttributeType))) ) != ((AttributeType *)NULL) )
		    	{
			    numMasterUpAtValues = 0 ;
			
			    for( curravs = masterUpavs ; curravs != NULLAV ; curravs = curravs -> avseq_next )
			    {
			    	if( (avcp = AttrV2pstr( &(curravs -> avseq_av) )) != NULLCP )
			    	{
				    fprintf( outfp, "%s & ", avcp ) ;
				    
			    	    if( (at = str2AttrT( avcp )) != NULLAttrT )
				    {
				    	*(masterUpAtValueAttrTlist + numMasterUpAtValues++) = at ;
				    }

				    free( avcp ) ;
				}
			    }

			    fprintf( outfp, "\n" ) ;

			    /* Now run through the attribute types retrieved from the Directory entry.. */
			    for( as = readAllEntryListp->ats ; as != NULLATTR ; as = as -> attr_link )
			    {
				attrTypeRejected = FALSE ;

				/* Check that the attribute type is not an RDN component.. */
/*				if( (lastEntryRDN = findLastRDN( readAllEntryListp->entryDN )) != NULLRDN &&
				    findAttrTypeRdnComp( as->attr_type, lastEntryRDN) != NULLRDN            )
				    attrTypeRejected = TRUE ;*/

				if( attrTypeRejected == FALSE )
				{
			    	    /* Check that the attribute type was not previously listed in tne "masterUpdatedAttrType".. */
				    for( i=0 ; i < numMasterUpAtValues ; i++ )
				    {
				    	if( AttrT_cmp( as->attr_type, *(masterUpAtValueAttrTlist + i)) == 0 )
				    	{
				    	    attrTypeRejected = TRUE ;
					    break ;
					}
				    }
				}
				    
				if( attrTypeRejected == FALSE )
				{
				    /* Now check the attribute type against the list of bulk load attributes
				       that should not appear in the "nonMasterUpdatedAttrType", also
				       reject "acl", "lastModifiedTime" and "lastModifiedBy" */

				    for( i=0 ; i < NUMOFREJECTATTRT ; i++ )
				    {
					if( rejectAttrsTable[ i ] != NULLAttrT &&
					    AttrT_cmp( as->attr_type, rejectAttrsTable[ i ]) == 0 )
				    	{
				    	    attrTypeRejected = TRUE ;
					    break ;
				    	}
				    }

				    if( attrTypeRejected == FALSE )
				    {
				    	/* We need to add this type as a value to "nonMasterUpdatedAttrType" */
					if( (avcp = AttrT2pstr( as->attr_type )) != NULLCP )
					{
					    if( (av = str2AttrV( avcp, nonMasterUpAtat-> oa_syntax )) != NULLAttrV )
					    {
						(void)insertAVintoAVS( &nonMasterUpavs, av ) ;
						AttrV_free( av ) ;

						nonMasterUpToBeUpdated = TRUE ;
					    }
					    
					    free( avcp ) ;
					}
				    }
				}
			    }

			    if( nonMasterUpToBeUpdated == TRUE )
			    {
			    	if( nonMasterUpAtInDirectory == TRUE )
				{
				    currModNodep = create_attr_mod( EM_REMOVEATTRIBUTE, nonMasterUpAtat, NULLAttrV ) ;
				    append_attr_mod( &baseModNodep, currModNodep ) ;
				}
			    
			    	/* Create the modification request for the "nonMasterUpdatedAttrType".. */
				currModNodep = create_attr_mod_from_AVS( EM_ADDATTRIBUTE, nonMasterUpAtat, nonMasterUpavs ) ;
				append_attr_mod( &baseModNodep, currModNodep ) ;
			    }			

			    /* Append on the delete "masterUpdatedAttrType" modification.. */
			    currModNodep = create_attr_mod( EM_REMOVEATTRIBUTE, masterUpAtat, NULLAttrV ) ;
			    append_attr_mod( &baseModNodep, currModNodep ) ;
			    
			    if( baseModNodep != NULLMOD &&
			        modifyDirectory( baseModNodep, nodeDN, &dserrorp ) != OK )
			    {
			    	if( (errorMsgcp = error2pstr( dserrorp )) != NULLCP )
			    	{
		    		    fprintf( outfp, "%s\n", errorMsgcp ) ;
		    		    free( errorMsgcp ) ;
			    	}
			    }

			    if( baseModNodep != NULLMOD )
			    {
			    	free_attr_mod_list( baseModNodep ) ; baseModNodep = NULLMOD ;
			    }

			    for( i=0 ; i < numMasterUpAtValues ; i++ )
			    {
				AttrT_free( *(masterUpAtValueAttrTlist + i) ) ;
			    }

			    free( (char *)masterUpAtValueAttrTlist ) ;
			}

			freeListpp( &readAllEntryListp ) ;
		    }
		    else 
		    {
			fprintf( outfp, "#ERROR read entry did not return any results for\n%s\n", DNcp ) ;
	 	    }
		}
	    }

	    if( nonMasterUpavs != NULLAV )
	    	avs_free( nonMasterUpavs ) ;
	    
	    freeListpp( &readMasterUpAtEntryListp ) ;
	}

	free( DNcp ) ;

	return( OK ) ;
}


void
processExchangeMasterUpAtforNonMasterUpAt( fatalInterrupt )
int	fatalInterrupt ;
{
FILE *	exchangefp ;

	if( fatalInterrupt == FALSE )
	{
	    /* Input the command parameters.. */
	    processMasterUpAtforNonMasterUpAtUpdates() ;

	    if( managementConfig.returnToMenu == TRUE )
	    	return ;

	    if( (exchangefp = fopen( managementConfig.masterUpAtToNonMasterLogFileNamecp , "w" )) == ((FILE *)NULL) )
	    	fprintf( stderr, "Attempt to open the output file failed\n" ) ;
	    else {
	    
		fprintf( exchangefp, "#\t\tREPORT OF THE ENTRIES THAT HAVE HAD THE masterUpdatedAttrType REPLACED/REMOVED\n#\t\t\
==============================================================================\n\n" );

	    	if( applyOperationToSubTreeNodes( exchangefp, managementConfig.masterUpAtToNonMasterSubTreeParentDN,
			exchangeMasterUpAtforNonMasterUpAt, FALSE,
			&(managementConfig.masterUpAtToNonMasterMode),
 			NULLAttrT, NULLAttrV ) == NOTOK )
		    fprintf( stderr, "masterUpdateAttrType adjustment failure for %s\n", managementConfig.masterUpAtToNonMasterSubTreeParentDNcp ) ;

	    	fclose( exchangefp ) ;
	    }
	}

	return ;
}
	
