.TL
Technical details and User Guide for
the bulk loading tools newdmdiff and bulkload.
.AU
Paul Barker & Kevin McCarthy
.AI
The Department of Computer Science,
University College London
.sp 3
.NH 1
Introduction
.PP
As part of University College London's contribution to the
.I Paradise
X.500 project two tools have been developed to facilitate the loading
and management of large quantities of data from a variety of data sources.
A high degree of automation has been incorporated into their 
specification to encourage directory
administrators to perform up-loads frequently. It is perceived
that the improved availability of ever larger quantities of up to date
information will lead to greater Directory utilisation.
.PP
The tools have been developed using \fIthe ISO Development Environment\fR,
making extensive usage of the \fIDSAP\fR library interface.
They support the standard X.500 attribute syntaxes together with those 
specific to \fIQUIPU\fR and \fIPP\fR.
.NH 2
Document Outline
.PP
There are four principle components to this document.
.IP \(bu
A description of the Bulk loading infrastructure.
.IP \(bu
A tutorial on the design and
loading of the infrastructure. Followed by information 
on uploading from multiple data sources.
.IP \(bu
Technical details on the tools' functionality.
.IP \(bu
An appendix contains the manual pages for the tools.
.NH 2
Design objectives
.IP \(bu
The bulk loading process shall facilitate a marked increase
in the quantity of data held, but not at the expense of sacrificing
data quality.
.IP \(bu
The tools shall facilitate bulk loading with minimal
intervention from the directory administrator.
.IP \(bu
The additional management information to be contained within
the Directory shall not be excessive. This prevents storage of
the data source that supplied each attribute value.
.IP \(bu
The tools must be able to handle several thousand entries per source update.
.bp
.RE
.NH 2
Overview of the functionality provided by the tools
.PP
The following tools are seen as a replacement to the
\fBdmtools\fR that are distributed as part of the \fIISODE-8.0\fR
package.
.PP
.B Newdmdiff
generates differences between two, potentially large, sets of
.I dmformat
data and is a replacement for
.B dmdiff.
In tests
.B newdmdiff
was shown to work significantly faster than
.B dmdiff.
More importantly search times increased only linearly with
each doubling in the sizes of the input files.
.PP
.B Bulkload
is responsible for the up load of either
.I dmformat
or
.I newdmformat
data from multiple sources, being controlled by the bulk management
infrastructure that is stored within the directory.
.PP
.B Bulkmanagement
\(dg facilitates correspondence checks between the directory
and the current data within the various sources.
In addition a
"down load" function is provided to write directory trees to
\fIdmformat\fR files. The tool also permits the collation and deletion of
expired entries.
.FS
\(dg This tools functionality has been built into \fBbulkload\fR
and is accessed when the \fI-m\fR command line option is specified.
.FE
.RE
.NH 2
Administrative Scenarios Envisaged
.PP
In all cases multiple data sources may be used to
update a given Directory sub-tree.
.SH
Subtree initialisation
.PP
Any data source applied shall be permitted to bulk load large number of
new entries to initialise DIT sub-trees. This facility is supported
by \fBbulkload\fR. Since the loading of very large amounts of data
is time consuming a non-interactive operational mode has been provided to
permit processing using either a background process or a night time batch job.
.RE
.KS
.SH
Day to day management
.PP
The \fBnewdmdiff\fR difference tool indicates those updates that need to
be applied to the directory on a day-by-day or week-by-week basis.
Taking differences between consecutive versions of the source data
minimises the number of updates that need to be applied and significantly
reduces the number of conflicts that can occur with existing data.
When loading, if conflicts occur with existing data,
\fBbulkload\fR can be used interactively to resolve them.
.KE
.KS
.SH
Directory Validation
.PP
This would be applied monthly or yearly to ensure that the directory
corresponds to the current contents of all the data sources.
Erroneously absent entries and values can then be added. Entries
that have not been checked against by any of the data sources that were used
in the validation process may be listed.
There should also be an opportunity to remove those entries that have
"expired". \fBBulkmanagement\fR is intended to fulfill these requirements.
.KE
.bp
.NH 1
The Bulk Load Management Infrastructure
.NH 2
Master data sources
.PP
The infrastructure defines a \fImaster source\fR as one that is a "preferred"
data source to maintain a specified attribute type within a given sub-tree.
.IP \(bu
Attribute types at each sub-tree level in the directory may have one or more
\fImaster sources\fR specified. These specifications are inherited at lower levels unless
other \fImaster source\fR specifications for the attribute types concerned are defined there.
.IP \(bu
Requests for a non-master source to add a value (or multiple values if added in a single operation)
will only be permitted if the directory attribute does not have any values.
.IP \(bu
When a \fImaster source\fR provides a value for a directory attribute for the first time
then, providing it is the first \fImaster source\fR to do so, any values already held by the attribute
will be cleared down before the value from the master source is added.
.IP \(bu
\fIMaster sources\fR are the only data sources permitted to remove attribute values.
.NH 2
Create and Delete data sources.
.PP
\fICreate sources\fR may be specified that designate those data sources that
are permitted to request the addition of directory entries.
.IP \(bu
Create source capability specifications are inherited by lower level sub-directories
unless one or more create sources are defined there.
.PP
\fIDelete sources\fR may be specified that designate those data sources that
are permitted to request the removal of directory entries.
.IP \(bu
Delete source capability specifications are inherited by lower level sub-directories
unless one or more delete sources  are defined there.
.NH 2
Algorithmic naming
.PP
Given the potential differences in entry naming between data sources each directory
entry shall contain one or more
.I
algorithmic name
.R
attribute values. These hold the names that the directory entry is known as
in each of the data sources that apply updates to it.
.NH 2
Entry expiry dates
.PP
Each entry may contain an expiry date which,
upon a request from the directory administrator, can be examined to determine
whether the entry should be removed.
.NH 2
An example of the management infrastructure
.PP
Let us now consider \fIfigure 1\fR.
.PP
Regarding attribute mastering we find that for {o=UCL}, "uclData" is the master
source for both \fIhomePhone\fR and \fIhomePostalAddress\fR.
.PP
The "telephoneNumber" attributes
within the {ou=CS} sub-tree are mastered by both the "csDept" and "uclPhoneData"
data sources. 
If the "dataSource=uclPhoneData" 
entry had not been defined within {ou=CS} the entry within {o=UCL} would not 
have been sufficient to make it a master source within {ou=CS}, since the lower level 
definition would have over ruled it.
.PP
The \fInonMasterUpdatedAttrType\fR attribute within "cn=Fred Bloggs"
indicates that a non-master source has added the \fIroomNumber\fR attribute 
value.
That is to say the \fInonMasterUpdatedAttrType\fR
attribute keeps track of all the attributes within a given entry that 
have had their values added by a non-master source.
The absence of \fItelephoneNumber\fR within \fInonMasterUpdatedAttrType\fR
prevents the \fItelephoneNumber\fR attribute being cleared down if
the other master source provides its first value.
.PP
If the optional \fIcreateSource\fR flag is absent or has been set to TRUE, the create entry
capability is enabled for the corresponding data source.
For {ou=CS} "csDept" is the only source with the create entry capability. Both
"csSystems" and "uclPhoneData" have the capability specifically disabled, whilst
"uclData"'s create capability can not be inherited due to the presence of
the "csDept" specification.
.PP
"UclData" is the delete source for all of UCL except the sub-tree {ou=CS}, as its 
inheritance is prevented by the lower level specification for the "csDept" data source.
.PP
The data entries such as that for "cn= Patricia Bloggs" contain
a list of \fIalgoName "data source name # algorithmic name within the data source"\fR
values. They indicate that the directory entry for Patricia Bloggs
should be updated from the record for "P H Bloggs" within the
"uclData" source and from the record for "Pat H Bloggs" within the
"uclPhone" source.
.PP
Both data entries have a value specified for the attribute \fIdeleteEntryAfter\fR
which may be read by the \fBbulkmanagement\fR
tool to check for expired entries.
.NH 2
Sub-directory search lists
.PP
It may well be the case that a given data source has a different organisational
structure than that used within the Directory.
Consider the following example which concerns updates to a "Registry" department which
contains three sections, namely Accounts, Exams and Records.
The Directory may well have an organisational unit for each of the three Registry sections.
However a telephone number data source may provide a single list of numbers
for the staff employed by all three sections.
Further more, the Directory may have the Registry sections grouped under
a Registry organisational unit or they may be held directly under
the college organisation. Sub-directory search lists may be used to allow
updates of the three Registry sections from the telephone data source.
.IP \(bu
A sub-directory search list specifies a sub-tree parent entry together
with a list of distinguished name extensions. These extensions, when
appended onto the parent entry's distinguished name, define the
distinguished names of the sub-directories that may be updated by a given data
source update.
.IP \(bu
If the entries within any of the organisational units that are immediately
below a parent entry are to be updated by a given data source then a special
\fIAll\fR case is permitted that does not require a sub-directory search list
entry to be defined.
.IP \(bu
The source purported names must be unique across the sub-directories that
are defined by the sub-directory search list.
.LP
Consider the following Directory structure:-
.TS
;
c c c c c.
		o= Brunel
ou=Accounts	ou=Art	ou=Exams	ou=Maths	ou=Records
.TE
.LP
When applying the Registry telephone data the required \fIbulkSubDirList\fR entry would be :-
.B
.TS
box;
l.
bulkSubDirList= RegistryList
bulkSubTreeDN= c=GB@o=Brunel
bulkSubTreeDNextension= ou=Accounts
bulkSubTreeDNextension= ou=Exams
bulkSubTreeDNextension= ou=Records
.TE
.R
.LP
Note: this example has only required single RDN distinguished name
extensions.
.PP
Whilst for Directory structure:-
.TS
;
c c c.
	o= Brunel
	ou=Registry
ou=Accounts	ou=Exams	ou=Records
.TE
the special case of "c=GB@o=Brunel@ou=Registry@bulkSubDirList=All"
would be used to apply the source's updates within Accounts, Exams
and Records.
.bp
.NH 1
Input file formats
.PP
Two formats are used by the bulk load tools. \fIDmformat\fR is used as the common
format that the various source formats must be converted into.
The \fInewdmformat\fR is used to apply incremental updates to the Directory.
.NH 2
.I Dmformat
.PP
This data format is based on \fIQUIPU's Entry Data Block\fR format, but also includes
.B source
and
.B rootedAt
statements.
The single
.B source
line names the data source.
The
.B rootedAt
statement defines the parent directory for the sub-directory to be updated
by the statements listed after it. Data supplied by a single source for an entire 
sub-tree may be included in the same file.
The \fBsource\fR and \fBrootedAt\fR statements and the RDN blocks are delimited 
by one or more blank lines.
.SH
An example of a
.I dmformat
file
.B
.TS
box;
l.
source= uclData

rootedAt=c=GB@o=University College London@ou=Computer Science

cn= Paul Barker
cn= P Barker
sn= Barker
uid= paulb
mail= P.Barker@cs.ucl.ac.uk
mhsORAddresses= /I=P/S=Barker/OU=cs/O=ucl/PRMD=UK.AC/ADMD=GOLD 400/C=GB/
userClass= csresstaff
roomNumber= G21
personalTitle= Mr
telephoneNumber= +44 71-380-7366
ou= Computer Science
objectClass= person & newPilotPerson & organizationalPerson & pilotObject & quipuObject & top & mhsUser

cn= David Goodman
cn= D Goodman
sn= Goodman
uid= dgoodman
mail= D.Goodman@cs.ucl.ac.uk
mhsORAddresses= /I=D/S=Goodman/OU=cs/O=ucl/PRMD=UK.AC/ADMD=GOLD 400/C=GB/
userClass= csadmin
roomNumber= G24
personalTitle= Mr
telephoneNumber= +44 71-380-7294
ou= Computer Science
objectClass= person & newPilotPerson & organizationalPerson & pilotObject & quipuObject & top & mhsUser

rootedAt=c=GB@o=University College London@ou=Mathematics

# ... and so on
.TE
.R
.bp
.NH 2
.I Newdmformat
.PP
This format is output by the difference tool.
Given the need to indicate when a data source is adding an
attribute type to an entry for the first time a further format has
been defined. This new format is the
.I dmformat
with a prepended
.B newdmformat
statement, together with usage of the following leading characters:-
.TS
box;
c|c|c
c|l|l.
Leading character	Indicates	Example
_
$	Add an entry	$cn= Paul Barker
_
!	Delete an entry	!cn= Paul Barker
_
*	Modify an RDN	cn= Pauline Barker % personalTitle=Miss
		*cn= Pauline Barker % personalTitle=Mrs
_
+	Add an attribute	+room= G22
_
-	Delete an attribute	-room
_
>	Add a value	>favouriteDrink= Whisky
_
<	Delete a value	<favouriteDrink= Water
.TE
.SH
An example of a
.I newdmformat
file
.B
.TS
box;
l.
newdmformat

source= ccData

rootedAt=c=GB@o=University College London@ou=Computer Science

cn= Andreas Antonopoulos
# Add a favouriteDrink attribute and then add a further value..
+favouriteDrink= Wine
>favouriteDrink= Spring water
# Add an additional object class attribute value..
>objectClass= bulkDataEntryDetails

cn= Fred Bloggs
# delete his favouriteDrink attribute..
-favouriteDrink
#remove a telephoneNumber attribute value..
<telephoneNumber= 0634-661122

# remove an entry..
!cn= John Bloggs

rootedAt=c=GB@o=University College London@ou=Mathematics

# modify an RDN (Note: no attribute updates can be listed
# in the same RDN block..)
cn= Frederica Bloggs % personalTitle= Miss
*cn= Frederica Bloggs % personalTitle= Mrs

# ... and so on
.TE
.R
.bp
.NH 1
Tutorial on setting up the bulk load infrastructure and applying updates
.PP
We shall now consider a short example to show how to define and load up
\fIdataSource\fR entries within a directory sub-tree.
The effects of the
\fIdataSource\fR entries on the bulk loading of data will then be considered.
The scenario is based on a Directory sub-tree which contains three
college departments represented as organisational units..
.TS
;
c c c.
	o=UCL	 
ou=CS	ou=Maths	ou=History
.TE
.NH 2
Available data sources
.PP
Seven data sources are available, csStaff, csStudent, csSystem (for
departmental computer account details), mathsData, uclComp (college wide
computer account details), uclPhone (college wide phone number information)
and uclData (college wide personal and locational details.)
The History department does not have any specific data sources.
.TS
box;
l|c s s s s s s
l|c|c|c|c|c|c|c.
Attribute	Attributes available from the data sources for the sub-tree
	_	_	_	_	_	_	_	
                 	csStaff	csStudent	csSystem	mathsData	uclComp	uclPhone	uclData
_
commonName	ou=CS	ou=CS	ou=CS	ou=Maths	o=UCL	o=UCL	o=UCL
surName	ou=CS	ou=CS		ou=Maths			o=UCL
telephoneNumber	ou=CS	ou=CS		ou=Maths		o=UCL	 
homePostalAddress	ou=CS	ou=CS		ou=Maths			o=UCL
mail			ou=CS		o=UCL	 	 
personalTitle	ou=CS	ou=CS		ou=Maths			o=UCL
roomNumber	ou=CS	ou=CS		ou=Maths			o=UCL
uid			ou=CS		o=UCL	 	 
.TE
.PP
Now that we know which attributes are available from which source
we must then determine, for those attributes that are available
from more than one source, which source is to be preferred.
It may be known that one source is more reliable and up to date
for certain attributes. It might even be the case that we have no reason to 
prefer one source to another.
For the csStaff and csStudent data, both are seen as preferred sources
for the same attributes, but in most cases there will
be no overlap in the Directory entries that they are required to update.
.TS
box;
l|c s s
l|c s s
c|c c c
c|c|c|c.
Attribute	Data source preferred to provide attribute value
	within departmental sub-tree
	_	_	_
	ou=CS	ou=Maths	ou=History
_
surName	csStaff or csStudent	mathsData	uclData
telephoneNumber	csStaff or csStudent	mathsData	uclPhone
homePostalAddress	uclData	uclData	uclData
mail	csSystem	uclComp	uclComp
personalTitle	csStaff or csStudent	mathsData	uclData
roomNumber	csStaff or csStudent	mathsData or uclData	uclData
uid	csSystem	uclComp	uclComp
.TE
.SH
Notes on the tables
.IP i)
From the tables we can see that for all three departments
uclData is the preferred source for \fIhomePostalAddress\fR as it is felt
that changes in the attribute will be reflected within this data set
before csStaff or csStudent (for the CS department) or mathsData (for the
Maths department).
.IP ii)
Whilst uclData also holds data for \fIroomNumber\fR
it is considered that the computer science departmental data sets
tend to hold more accurate room number data. For the Maths department
the mathsData and the uclData sources are considered equivalent when
it comes to providing \fIroomNumber\fR data.
.NH 2
Determining the create and delete source capabilities
.PP
The "csSystem" data source will need to create account entries
(as opposed to personal entries) within {ou=CS} and so it must have a create 
entry capability. "UclComp" is only permitted to add data to existing
entries.
"csStaff", "csStudents" and "mathsData" need the create capability within
their respective sub-directories. The History department's entries will
be created from the "uclData" source. "UclData" will be permitted to
create entries within {ou=Maths}.
The data source "uclPhone" is not permitted to create entries as it is
intended to provide updates for existing entries.
.PP
Regarding the delete entry capability all sources having the create
capability shall also have the delete capability, with one exception.
Within {ou=Maths} we do not want to give "uclData" the ability to
remove entries.
.NH 2
Defining the required \fIdataSource\fB entries\fR
.PP
Given the specifications for the preferred sources to supply each attribute 
and for the create and delete entry capabilities we can now construct
the \fIdataSource\fR entries that we require.
.SH
o=UCL
.PP
.TS
box;

l|l|l|l.
\fIdataSource\fR	\fImasters\fR	\fIcreateSource\fR	\fIdeleteSource\fR
=
uclComp	mail, uid	FALSE	FALSE
_
uclPhone	telephoneNumber	FALSE	FALSE
_
uclData	surName,	TRUE	TRUE
	homePostalAddress,	
	roomNumber,	
	personalTitle	
.TE
.SH
ou=CS
.PP
.TS
box;
l|l|l|l.
\fIdataSource\fR	\fImasters\fR	\fIcreateSource\fR	\fIdeleteSource\fR
=
csStaff	surName, telephoneNumber	TRUE	TRUE
 	personalTitle, roomNumber	
_
csStudent	surName, telephoneNumber	TRUE	TRUE
 	personalTitle, roomNumber	
_
csSystem	mail, uid	TRUE	TRUE
.TE
.SH
ou=Maths
.PP
.TS
box;
l|l|l|l.
\fIdataSource\fR	\fImasters\fR	\fIcreateSource\fR	\fIdeleteSource\fR
=
mathsData	surName, telephoneNumber	TRUE	TRUE
 	personalTitle, roomNumber	
_
uclData	roomNumber	TRUE	FALSE
.TE
.LP
Note: the specification of uclData as master source for \fIroomNumber\fR
must be defined under {ou=Maths}, despite being already defined under
{o=UCL},
since inheritance is prevented by the presence of the mathsData mastership
specification for this attribute.
.SH
ou=History
.PP
No additional \fIdataSource\fR entries are required. Any master, create or
delete source specifications, which are defined under {o=UCL}, will be inherited.
.NH 3
Other issues in defining \fIdataSource\fB entries\fR
.PP
For \fIQUIPU\fR implementations we must consider the presence
of \fItreeStructure\fR attributes. It may well be the case that the
entry for {o=UCL} contains a \fItreeStructure\fR attribute, which
would require us to add in the object classes used within the bulk loading
infrastructure. These classes are \fIbulkDataSourceEntry\fR,
\fIbulkDataSourceDetails\fR, \fIbulkDataEntryDetails\fR and
\fIbulkSubDirListClass\fR.
.PP
Regarding the access control requirements to be applied to
the \fIdataSource\fR
entries, it is considered that the default settings provided by the
\fIQUIPU\fR DSA are acceptable. There seems to be no reason why read access
should be restricted.
.PP
When planning the \fIdataSource\fR entries required it is essential that the
current settings of the configuration statements denoted (I), (II), (III),
(IV) and (V) below are reviewed. Since the settings will be required
for \fBbulkload\fR updates in the future any changes should be made to
the \fIbulkloadtailor\fR file itself and not to the configuration buffer
within \fBbulkload\fR which only has effect for the duration of the tools
execution.
.IP I)
"Highest directory level to search for dataSource entries".
For our example we would require \fIdataSource\fR entries below {c=GB@o=UCL}
to form our bulk load infrastructure and so the highest search level should be
set to "2".
.IP II)
"All sources are masters for".
The attribute types that are to be mastered by all data sources
should include \fIobjectClass\fR and \fItreeStructure\fR, together with
\fImasters\fR, \fIcreateSource\fR and \fIdeleteSource\fR if \fBbulkload\fR is to update
(as opposed to merely create) \fIdataSource\fR entries.
.IP III)
"When no master specified for a type all sources are masters". Ensure
this statement is set to "NO".
.IP IV)
"All sources are create sources : YES".
Specify that all data sources are permitted to create entries
whether there is a \fIdataSource\fR entry that permits them to or not.
We require that this should be set to "NO" since we have specified all the sources
that are to be permitted to create entries.
.IP V)
"Existing attr value may be cleared down if master source applies 1st value : NO".
Would prevent an attribute's values being cleared down when a master data
source adds a value for the first time (as recorded by the
\fInonMasterUpdatedAttrType\fR attribute.)
.KS
.NH 3
Using bulkload to load the \fIdataSource\fB entries\fR
.PP
Should a \fItreeStructure\fR attribute be present in the entry for {ou=CS}
then we can add the bulk load object classes using the following statements:-
.B
.TS
box;
l.
newdmformat

source=bulk

rootedAt=c=GB@o=University College London

o=Computer Science
>treeStructure= bulkDataSourceEntry
>treeStructure= bulkDataSourceDetails
>treeStructure= bulkDataEntryDetails
>treeStructure= bulkSubDirListClass
.TE
.R
.LP
.KE
.RE
.PP
When adding an entry with a \fIdataSource\fR attribute in the RDN
the \fIbulkDataSourceDetails\fR object class attribute will be added
automatically (if absent.) Given that we wish to create new entries,
either the \fIdmformat\fR or the \fInewdmformat\fR may be used.
All sources are automatically permitted to create \fIdataSource\fR
entries.
The following \fIdmformat\fR listing can be used to load up the
the \fIdataSource\fR entries that have been specified above.
.B
.TS
box;
l.
source=bulk

rootedAt=c=GB@o=University College London

dataSource= uclComp
masters= mail & uid
deleteSource= FALSE
createSource= FALSE

dataSource= uclPhone
masters= telephoneNumber
deleteSource= FALSE
createSource= FALSE

dataSource= uclData
masters= surName & homePostalAddress & roomNumber & personalTitle
deleteSource= TRUE
createSource= TRUE

rootedAt=c=GB@o=University College London@ou=Computer Science

dataSource= csStaff
masters= surName & telephoneNumber & personalTitle & roomNumber
deleteSource= TRUE
createSource= TRUE

dataSource= csStudent
masters= surName & telephoneNumber & personalTitle & roomNumber
deleteSource= TRUE
createSource= TRUE

dataSource= csSystem
masters= mail & uid
deleteSource= TRUE
createSource= TRUE

rootedAt=c=GB@o=University College London@ou=Mathematics

dataSource= mathsData
masters= surName & telephoneNumber & personalTitle & roomNumber
deleteSource= TRUE
createSource= TRUE

dataSource= uclData
masters= roomNumber
deleteSource= FALSE
createSource= TRUE
.TE
.R
.PP
If these are held in the file "bulkDataSourceFile" they can be interactively loaded using:-
.IP
\fBbulkload\fR bulkDataSourceFile -o i
.NH 2
Usage of sub-directory search lists
.PP
Unfortunately the uclComp data source is not structured
in the same way as the Directory, since the data for
the two science departments is combined together.
We thus need to specify a sub-directory search list for uclComp.
.B
.TS
box;
l.
source=bulk

rootedAt=c=GB@o=University College London
# .. or even a special sub-directory for storing search lists !

bulkSubDirList= uclCompScienceList
bulkSubTreeDN= c=GB@o=University College London
bulkSubTreeDNextension= ou=Maths
bulkSubTreeDNextension= ou=CS
.TE
.PP
If this is held in the file "bulkUclCompSubDirList" it can be interactively loaded using:-
.IP
\fBbulkload\fR bulkUclCompSubDirList -o i
.R
.LP
Note: again we have only needed to use single RDN distinguished name
extensions.
.NH 2
Data loading with bulkload
.PP
Now that the required \fIdataSource\fR entries have been loaded into the
Directory we can consider the issues involved in actually loading and
maintaining the data to be provided by the various data sources.
.NH 3
Data formatting
.PP
It may well be the case that the available data sources are represented
in a number of different formats.
These formats must first be converted into the \fIdmformat\fR. \fIAwk\fR
or \fIperl\fR could be used to carry out the required conversions.
.PP
Where ever possible the updates should be structured into the
organisational units that are used in the Directory.
This permits the usage of \fBrootedAt\fR statements for specific
Directory units.
When this is not feasible sub-directory search lists would need to be used.
.PP
When specifying the object classes that need to be added within a create
entry request the object class insertion lists, that are defined within
the configuration file, may be of assistance.
.PP
Continuing with the current example any updates from the data sources
csStaff, csStudent, csSystem, mathsData, uclPhone
and uclData would be grouped under the organisational units that they
apply to and as specified in the accompanying \fBrootedAt\fR statements.
The uclComp data would provide two blocks of updates, one for the
History department and one for the two science departments
.NH 3
Data loading
.PP
Those sources that provide the mandatory attributes for entry creation should
be loaded into the Directory first, such as departmental data sets.
Other, secondary sources, such as those that merely provide telephone numbers
or email addresses, may then be applied. When applying a secondary source,
\fBbulkload's\fR "-e" option should be used, as this notifies the tool
that the data is meant for existing directory entries and so
the matching rules used to pair off a source entry's purported name and
a Directory RDN can be relaxed. If the "-e" option is not used
operator intervention would be required to pair off entries
if an exact match between a source purported name and a Directory entry's RDN
could not be found.
Once a data source has been applied to a Directory entry an
\fIalgoName\fR attribute is added to it, thus ensuring that on subsequent
updates the source purported name can be paired off with the required
Directory entry automatically.
.PP
When a new copy of a data source becomes available \fBnewdmdiff\fR
should be used to produce a list of differences, in the \fInewdmformat\fR,
which can then be applied to the Directory. 
.PP
Non-master sources (with respect to a given attribute type in a given
Directory sub-tree) are permitted to add a value (or more than one if added
in a single operation) to those attributes that do not currently have any values from a master source.
If a value had already been
provided by a master source, an error condition would arise which is handled as specified
by the configuration statements "Add value when not master source" and
"Add attribute when not master source".
.NH 4
Data loading using sub-directory lists
.PP
The main difference is that the \fBrootedAt\fR statements
include a \fIbulkSubDirList\fR RDN component.
One of the two uclComp \fBrootedAt\fR statements would specify
a list component :-
.B
.TS
box;
l.
source=uclComp

rootedAt= c=GB@o=University College London@bulkSubDirList=uclCompScienceList

# updates for the uclComp maths and computer science data..

rootedAt= c=GB@o=University College London@ou=History

# updates for the uclComp art departmental data..
.TE
.R
.PP
If we had needed to utilise the \fIAll\fR sub-directories special case
the \fBrootedAt\fR would have been:-
.B
.TS
box;
l.
rootedAt= c=GB@o=University College London@bulkSubDirList= All

# updates to apply within the organisational units immediately below
# c=GB@o=University College London i.e. CS, Maths and History.

.TE
.R
.bp
.NH 4
Sample upload
.PP
The data set csStaff is to be loaded for the first time.
.B
.TS
box;
l.
source= csStaff

rootedAt=c=GB@o=University College London@ou=Computer Science

cn= Graham Knight
cn= G Knight
sn= Knight
telephoneNumber= +44 71-380-7366
homePostalAddress= 29 Databus Drive $ ChipVille $ London
personalTitle= Mr
roomNumber= G21
objectClass= person & newPilotPerson & organizationalPerson & pilotObject & quipuObject

cn= Paul Barker
cn= P F Barker
cn= P Barker
sn= Barker
telephoneNumber= +44 71-380-7366
homePostalAddress= 500 Directory Avenue $ Datatown $ London
personalTitle= Mr
objectClass= person & newPilotPerson & organizationalPerson & pilotObject & quipuObject
.TE
.R
.PP
If this was in the file csStaffData we can interactively load it into the Directory using :-
.IP
\fBbulkload\fR csStaffData -o i
.PP
Next we wish to apply the uclData details to supply any missing
\fIpersonalTitle\fR or \fIroomNumber\fR values and to provide the authoritative
\fIhomePostalAddress\fR details.
.B
.TS
box;
l.
source= uclData

rootedAt=c=GB@o=University College London@ou=Computer Science

cn= G Knight
sn= Knight
homePostalAddress= 92 ISDN Drive $ ChipVille $ London
personalTitle= Mr
roomNumber= G21

cn= P F Barker
sn= Barker
homePostalAddress= 500 Directory Avenue $ Datatown $ London
personalTitle= Mr
roomNumber= G21
.TE
.R
.bp
Since we know we are updating existing entries and given that the purported names
"G Knight" and "P F Barker" do not exactly match the \fIcommonName\fR
value for the existing entries, by using the "-e" flag we can avoid having to confirm 
the purported name to entry matches:-
.IP
\fBbulkload\fR uclCSData -o i -e
.LP
During loading, since there is only one existing Directory entry
with the same initials and surname for each of the purported names
in the update, the pairing off may be done automatically.
Being the first master source to supply values for \fIhomePostalAddress\fR
the value for the entry "cn= Graham Knight" will now be updated.
Additionally, the entry for "cn=Paul Barker" will receive a value for \fIroomNumber\fR
since no value was previously held, despite that fact that uclData is not the master
source for the attribute type within {ou=CS}.
.PP
A subsequent copy of csStaff now becomes available. We first apply \fBnewdmdiff\fR
to obtain the output file:-
.IP
\fBnewdmdiff\fR csStaffData csStaffData2 > uclCSDiff
.PP
Which produces:-
.B
.TS
box;
l.
newdmformat

source= csStaff

rootedAt=c=GB@o=University College London@ou=Computer Science

cn= Graham Knight
>roomNumber= G22
<roomNumber= G21
.TE
.R
This may then be applied:-
.IP
\fBbulkload\fR uclCSDiff -o i
.LP
Since csStaff masters \fIroomNumber\fR the operation will succeed and Graham Knight
gets a new room number.
.bp
.NH 1
Technical details of the operation of the tools
.NH 2
.B Newdmdiff
.PP
To ensure the correct matching between attribute types and values
this difference tool uses the
.I QUIPU
.I AttrT_cmp()
and
.I AttrV_cmp()
procedures throughout.
Given the tools' usage of 
.I AttributeTypes 
it requires access to the
.I ISODE
etc files.
.PP
Although the "&" multiple attribute value separator may be
present in the
.I dmformat
input files, the
.I newdmformat
output files list each value as "attribute type= value".
.PP
The tool supports the distinction between attribute addition/deletion
("+"/"-") and attribute value addition/deletion (">"/"<") as
this assists the parsing to be carried out by the
.B bulkload
tool and allows for the special actions that need to be undertaken when
a master source adds an attribute value for the first time.
.NH 2
.B Bulkload
.PP
.NH 3
The Configuration File
.PP
In addition to the system level details such as the address of the DSA
to be updated and the user name, a fairly large number of parameters
have been included to ease the operation of these tools.
.NH 4
DSA access parameters
.DS L
dsa_address:   Internet=128.16.8.27+15437
.DE
.IP
This is the address of the access point DSA. It should be noted that
\fBbulkload\fR will re-bind automatically upon a DAP referral on modification operations.
.LP
username:@c=GB@o=University College London@ou=Computer Science@cn=uclManager
.IP
This is the user name which the DUA binds to the directory with.
.LP
password: XyhJ
.IP
The password to be used when binding. \fBPlease ensure that your tailor file
is suitably protected\fR.
.LP
Note: if the "username" or "password" statements are absent from the tailor
file then the user's ".quipurc" file will be searched. Alternatively
they may be input from the command line using the "-u" and "-p" options.
.NH 4
Display parameter
.LP
Display config at start : YES
.IP
The switch must be either "YES" or "NO".
Indicate whether you require the bulk load specific configuration details
to be displayed for editing/review upon initial execution of
.B bulkload.
.NH 4
\fIdataSource\fR \fBscoping, master and create source determination\fR
.PP
In an effort to ease the administrative work load in setting up the
master source details, within the bulk load management infrastructure, three methods
have been provided. When \fBbulkload\fR determines whether a data source is a master source 
for an update to a particular attribute type instance the results of these methods 
are examined  in the order they are listed here :-
.IP (1)
The "All sources are masters for" configuration statements may be used
to indicate that all data sources are considered as master sources for
the specified attribute types.
.IP (2)
\fIdataSource\fR entries define the attribute types to be mastered
by each data source,
(providing they are the lowest level mastership definitions within the 
search scope defined by the \fBrootedAt\fR
distinguished name and the "Highest directory level to search for dataSource entries"
configuration statement.) Note: \fIdataSource\fR entries also indicate
whether the source is a \fIdelete source\fR or not.
.IP (3)
Finally, the "When no master specified for a type all sources are masters"
configuration statement can be used to give blanket mastership to all
attribute types that do not have a master source specified by either methods (1)
or (2).
.LP
Highest directory level to search for dataSource entries : 2
.IP
The accepted values are 1 to 9.
The
.I dataSource
entries that affect the operation of
.B bulkload
are retrieved from the directory from the level specified here down to
that indicated in the
.B rootedAt
statement being processed at the time.
As an example, for the distinguished name within the following line:-
.IP
rootedAt= c=GB@o=University College London@ou=Computer Science
.IP
and a highest search level of 1 then the
.I dataSource
entries within :-
.RS
.IP i)
c=GB
.IP ii)
c=GB@o=University College London
.IP iii)
c=GB@o=University College London@ou=Computer Science
.RE
.IP
will be retrieved.
.IP
Whilst for a highest search level of 3, only the \fIdataSource\fR
entries within :-
.RS
.IP i)
c=GB@o=University College London@ou=Computer Science
.RE
.IP
shall be retrieved.
.KS
.LP
.TS
;
l c l.
All sources are masters for	:	objectClass
All sources are masters for	:	favouriteDrink
.TE
.IP
The right-hand side token must be an attribute type.
It may well be the case that for attribute types such as \fIobjectClass\fR
all data sources may need to amend them to avoid object class constraint
violations when updating attribute types that they are masters of according
to a \fIdataSource\fR entry. For other types e.g. \fIfavouriteDrink\fR the 
administrator may not need to place any restrictions on which source can provide updates.
.KE
.KS
.LP
Existing attr value may be cleared if master source applies 1st value : YES
.IP
The switch must be either "YES" or "NO".
Any existing values for an attribute may be cleared down when a master
source adds a value to the attribute for the first time. (Whether it
is the first time or not is indicated by the absence of the attribute's type
as a value within the entry's \fInonMasterUpdatedAttrType\fR attribute.)
.KE
.KS
.LP
When no master specified for a type all sources are masters : NO
.IP
The switch must be either "YES" or "NO".
If the "dataSource" entries found within the scope defined by the
"Highest directory level to search" and the Distinguished Name within a 
.B rootedAt
statement do not specify a master source for a given
attribute type then this configuration statement determines whether all
data sources are treated as the attribute type's master source
(or none of them are). Wherever possible the "All source are masters for"
statement should be used instead, as it is more specific.
.KE
.KS
.LP
All sources are create sources : YES
.IP
The switch must be either "YES" or "NO".
Specify that all data sources are permitted to create entries
whether there is a \fIdataSource\fR entry that permits them to or not.
.IP
Note: all sources are treated as being create sources when creating
the \fIdataSource\fR entries themselves.
.KE
.bp
.NH 4
Routing/handling upon an error condition
.LP
.TS
;
l c l.
Add value that already exists	:	NOACTION
Add attribute that already exists	:	NOACTION
Add value when not master source	:	PROCESS
Add attribute when not master source	:	PROCESS
Delete value when not master source	:	PROCESS
Delete attribute when not master source	:	PROCESS
Delete value that does not exist	:	NOACTION
Delete attribute that does not exist	:	NOACTION
Create entry when not create source	:	PROCESS
Delete entry when not delete source	:	PROCESS
.TE
.IP
The accepted values are "NOACTION", "PROBLEM" or "PROCESS".
These statements define the action to be taken when one of the
above errors is detected. "NOACTION" and "PROBLEM" will cause
the
.I dmformat
or
.I newdmformat
line/RDN block that caused the error to be transferred to the corresponding
".noAction" or ".PROBLEM" file. When in non-interactive
mode, "PROCESS" results in output to the ".PROBLEM" file, but when in interactive
mode the line(s) in error will be presented for editing, with the error message,
for correction or routing to an output file.
.IP
It should be noted that checks for "Add value/attribute that already exists"
error conditions are applied before those for "Add value/attribute when not master source".
Whilst checks for "Delete value/attribute when not master source" error conditions 
are applied before checks for "Delete value/attribute that does not exist".
.LP
Auto convert add entry to add attributes if entry already exists : YES
.IP
The switch must be either "YES" or "NO".
It may well be the case that the same entry is requested to be added by
more than one data source e.g. an email source specifies a request for
entry creation after a departmental data source has already caused the
entries to be created. If the entry already exists we can request that
an attempt is made to add the attributes contained in the new add entry
request.
Auto conversion from entry creation to attribute addition will only come
into operation if:-
.IP a)
An algorithmic name match is found (see 5.2.2.1).
.IP b)
The purported name exactly matches an existing entry (see 5.2.2.2).
.IP c)
When the "-e" option is used and only a single directory entry
has the same initials and surname as the purported name, with at least
a possible match of title e.g. "Dr" to "Mr".
.NH 4
Management infrastructure matching and entry expiry
.LP
.TS
;
l c l.
MALE	:	Mr
MALE	:	Sir
FEMALE	:	Miss
FEMALE	:	Ms
FEMALE	:	Mrs
FEMALE	:	Lady
HUMAN	:	Dr
HUMAN	:	Prof
NONHUMAN	:	Daemon
.TE
.IP
These statements define lists of personal titles that are used to find
unambiguous matches during searches for existing entries. The "HUMAN"
grouping holds those titles that are not gender specific, such as academic
titles. Further details
can be found in the section on entry matching rules (see 5.2.2.)
.bp
.LP
.TS
;
l c l c l.
checkAttrTypeForDeleteEntry	:	userClass	 	 
periodBeforeDeletion	:	3rdyrcs	:	14
periodBeforeDeletion	:	csacstaff	:	NOLIMIT
.TE
.IP
These statements are optional.
.IP
The numerical value within
.I periodBeforeDeletion
is in months from the current date.
If these statements are present when an entry is being added to the directory,
.B bulkload
is required to automatically add a \fIdeleteEntryAfter\fR attribute value.
The UTC time for deletion depends on the value within the attribute specified by the
.I checkAttrTypeForDeleteEntry
statement i.e.
.I userClass.
.IP
As an example if
.B bulkload
adds an entry for "cn= Fred Bloggs" having a
.I userClass
of "3rdyrcs" then the
.I deleteEntryAfter
attribute will hold a UTC time 14 months after the current date.
For a "NOLIMIT"
.I periodBeforeDeletion
user class no
.I deleteEntryAfter
attribute will be added. 
.KS
.NH 4
Object class list expansion
.LP
.TS
;
l c l.
autoInsertOnAttrT	:	objectClass
autoInsertOnAttrV	:	person
autoInsertObjCls	:	organizationaPerson
autoInsertObjCls	:	newPilotPerson
autoInsertObjCls	:	pilotObject
autoInsertObjCls	:	quipuObject
		
autoInsertOnAttrT	:	bulkEntryClass
autoInsertOnAttrV	:	personClass
autoInsertObjCls	:	person
autoInsertObjCls	:	organizationaPerson
autoInsertObjCls	:	newPilotPerson
autoInsertObjCls	:	pilotObject
autoInsertObjCls	:	quipuObject
.TE
.IP
When \fBbulkload\fR constructs a create entry request, the entry's attributes
are checked for the attribute type/value pairings specified by the
"autoInsertOnAttrT" and "autoInsertOnAttrV" statements. If found
the list of \fIobjectClass\fR attributes that are specified by the
"autoInsertObjCls" statements for each of the matched type/value pairings
will be inserted automatically.
.KE
.KS
.NH 4
Bulkmanagement configuration details
.LP
default DN : "c=GB@o=University College London@ou=Computer Science"
.IP
Various management options present the contents of the default
distinguished name during parameter input.
.KE
.bp
.NH 3
Entry matching rules.
.PP
When determining a match between an entry being added or updated
and an existing entry \fBbulkload\fR runs through up to three steps.
The searches are applied to the directory entries having the parent entry 
defined by the distinguished name in the
\fBrootedAt\fR statement. The way the results of the matches performed are treated
differs between the add entry and the update entry cases - i.e.
an update requires that a corresponding directory entry already
exists whilst an add entry requires that is does not.
.NH 4
Algorithmic name search
.PP
Perform an algorithmic name search for a directory entry holding an
.I algoName
attribute value containing the current data source name and the RDN
of the entry that is to be added/updated.
If no exact match occurs try the second step.
.NH 4
Read entry
.PP
Perform a read entry operation for the RDN of the entry to be added/updated.
If the read does retrieve an entry then add the missing
.I algoName
attribute and also the \fIbulkDataEntryDetails \fR object class if missing, 
else try the third step.
.NH 4
Checks for unambiguous matches
.PP
Perform unambiguous entry matches.
For an add entry case
.B bulkload
evaluates each directory entry's
.I commonName
and
.I personalTitle
attributes to ensure that existing entries can be disambiguated from the
RDN of the entry to be created. The matching allows for the possibility
of titles appearing in the
.I commonName
attribute value. Personal titles are considered to be ambiguous (or not)
as indicated in the following table and as specified in the
configuration file.
.TS
allbox;
l c c c c.
	Male	Female	Human	Non-human
Male	ambiguous	unambiguous	ambiguous	unambiguous
Female	unambiguous	ambiguous	ambiguous	unambiguous
Human	ambiguous	ambiguous	ambiguous	unambiguous
Non-human	unambiguous	unambiguous	unambiguous	ambiguous
.TE
.PP
Initials and forenames within existing entries are sufficient
to disambiguate an existing entry with the entry to be added.
When matching initials only the number held within the entry
having the least number are checked e.g. "A B Bloggs" can not
be disambiguated from "A B C Bloggs" whilst "A B D Bloggs" can.
When comparing forenames to initials only the forename's initial
is used.
.PP
When updating an existing entry in interactive mode
.B bulkload
provides a list of existing entries having a
.I commonName
match with "*<first initial>*<space><surname>*" (where '*' denotes a wild
card) so that the user may specify the one (if any) that the updates are
to be applied to. Upon selection, an \fIalgoName\fR attribute and the
\fIbulkDataEntryDetails\fR object class (if there is no bulk data object
class present) will be added to the directory entry.
.bp
.NH 3
Updating Rules
.NH 4
Updates to the bulk management attributes
.SH
Updates to \fIalgoName \fR
.PP
When adding an entry the \fIalgoName \fR attribute will be included
automatically. It will be based on the data source name from the
.B source
statement and the entry's RDN. If a sub-directory search list
is being used then the search list's name (e.g.
\fIbulkSubDirList=uclCompScienceList"\fR) will also be included in the
algorithmic name e.g. 
.PP
\fIalgoName\fR= "csData # bulkSubDirList=uclCompScienceList % cn=Paul Barker"
.PP
During RDN modification requests the existing \fIalgoName \fR attribute
will be replaced with one derived from the new RDN.
.PP
The attribute will also be added if a match occurs within
the "Read entry" matching phase (see 5.2.2.2) and the
"Checks for unambiguous matches" phase for existing entries when 
in interactive mode (see 5.2.2.3).
.SH
Updates to \fIobjectClass \fR
.PP
Under the cases to be listed a \fIbulkDataEntryDetails \fR object class
shall be added to the directory entry if the entry does not already
contain a \fIbulkDataSourceEntry\fR,  \fIbulkDataSourceDetails \fR,
\fIbulkDataEntryDetails\fR or \fIbulkSubDirListClass\fR object class.
.PP
The object class insertion list facility leads to the incorporation
of object class lists (if specified) during the construction of DSAP 
create entry requests.
.PP
During entry addition the \fIbulkDataEntryDetails\fR object class
will be added automatically.
This class may also be added within
the "Read entry" matching phase (see 5.2.2.2) and the "Unambiguous"
matching phase for existing entries (see 5.2.2.3).
.SH
Updates to \fIdeleteEntryAfter \fR
.PP
When adding an entry the \fIdeleteEntryAfter \fR attribute will be included
automatically. The UTC time specified will be derived by adding
the period before deletion to the current system time. The period before
deletion is determined from the value of the the entry's attribute type that
was defined by the configuration statement
\fIcheckAttrTypeForDeleteEntry \fR and the corresponding
\fIperiodBeforeDeletion \fR value.
.SH
Updates to \fInonMasterUpdatedAttrType\fR
.PP
Before attribute updates are processed for a given RDN the
\fInonMasterUpdatedAttrType\fR values are retrieved for the entry.
A check is made to ensure that the corresponding attributes
still have values, if they do not the respective 
\fInonMasterUpdatedAttrType\fR value will be removed.
.PP
\fInonMasterUpdatedAttrType \fR will be added automatically
during entry addition.
When an add attribute request succeeds for a master data source,
the entry's \fInonMasterUpdatedAttrType \fR attribute will have the
newly added attribute's type removed as a value (if present).
.PP
The entry's \fInonMasterUpdatedAttrType \fR will have an attribute type
value removed if a remove attribute request succeeds. A value will be
added if an add attribute succeeds for a non-master data source.
.bp
.NH 4
Updates to non-bulk management attributes
.SH
Entry deletion
.PP
Delete entry requests will be rejected if the data source requesting it
is not the delete source.
.SH
Delete attribute or delete value requests
.PP
Attribute deletion may only be performed by a master source.
If a delete value request is issued that would remove the last
Directory entry value for the respective attribute type then
a delete attribute request will be substituted.
.SH
Attribute addition
.PP
If the data source is a master source for the attribute type being updated
and the entry's
\fInonMasterUpdatedAttrType\fR has the attribute type listed
as a value then the attribute will initially be removed (to clear any
values added by a non-master source.)
.PP
If the data source is a master source and the entry does not
currently hold any values for the given attribute type
an add attribute request is performed. If values are already
present an add value operation is applied.
.KS
.SH
Attribute value addition
.PP
If the data source is a master source for the attribute type being updated
then an add value request is performed.
.PP
If the data source is NOT a master source then an \fBadd attribute\fR (as opposed
to an an add attribute value) request is initiated which will fail
if the attribute type already has any values.
.KE
.sp 3
.TS
box;
c s s s
c|c|c|c.
\fIMaster Source\fR Update Rules
_
Update	Data source is a	\fInonMasterUpdatedAttrType\fR	update
requested	master source	has type already listed	attempted
_
Add Attribute	No	 	Add Attribute which
 	 	 	will fail if the directory
 	 	 	already holds any value(s)
	_	_	_
  "       "	Yes	Yes	Add Attribute after
 	 	 	a Remove Attribute has
 	 	 	cleared existing values
	_	_	_
  "       "	Yes	No	Add attribute (or add value
         			if values already present)
_
Add Value	No	 	\fBAdd Attribute\fR which
 	 	 	will fail if the directory
 	 	 	already holds any value(s)
	_	_	_
 "     "	Yes	 	Add Value
_
Delete Attribute	No	 	NONE
	_	_	_
 "     "	Yes	 	Remove Attribute
_
Delete Value	No	 	NONE
	_	_	_
 "     "	Yes	 	Remove Value (unless last
      		 	value, then Remove Attribute)
.TE
.bp
.SH
Appendix I : Manual Pages
.SH
NAME
.ce
\fB\s+5newdmdiff\s-5\fR
.SH
SYNOPSIS
.DS L
\fBnewdmdiff\fR  [ -so | -s ] old\fIdmformat\fRfile  new\fIdmformat\fRfile
.DE
.SH
EXAMPLE USAGE
.TS
;
l l.
\fBnewdmdiff\fR	janUCLCSdata febUCLCSdata > UCLCSdiff
.TE
.SH
DESCRIPTION
.PP
\fBNewdmdiff\fR will parse both input files and produce a \fInewdmformat\fR 
output of the differences.
In the input file all the entries for a given parent Distinguished Name
should be grouped under a single parent \fBrootedAt\fR.
In the output file any attribute additions will be listed before attribute deletions.
.PP
The tool will not replace delete value requests that would remove all of an attribute's
values, that are listed by a given data source for an entry, with a delete attribute
request. This permits more than one source to maintain the same attribute type (i.e.
to act as \fImaster\fR sources.)
.PP
If the tool is being used to differ secondary data sources, i.e. those that
merely supply additional attributes (as opposed to creating entries), then when
an entry is found to be missing from the new file the user may require that
attribute deletion requests are output instead of entry deletion requests.
.SH
OPTIONS
.LP
-so
.IP
For secondary data sources when an entry RDN is missing from the
new\fIdmformat\fRfile, output delete value requests instead of
marking the entry for deletion. \fIObjectClass\fR attributes
shall be included.
.LP
-s
.IP
For secondary data sources when an entry RDN is missing from the
new\fIdmformat\fRfile, output delete value requests instead of
marking the entry for deletion. \fIObjectClass\fR attributes
shall be excluded.
.bp
.SH
NAME
.ce
\fB\s+5bulkload (bulkmanagement)\s-5\fR
.SH
SYNOPSIS
.TS
;
l l.
\fBbulkload\fR	[ \fIdmformat\fRfile | \fInewdmformat\fRfile ] ]
           	[ -o [ Interactive | Non-interactive ] ]
           	[ -m ]
         	[ -t tailorfile ]
        	[ -e ]
        	[ -p password ]
        	[ -u "Directory user name" ]
        	[ -d ]
		[ -m ]
.TE
.SH
EXAMPLE USAGE
.DS L
\fBbulkload\fR  UCLCSdiff  -o i  -t mybulkloadtailor  -e
\fBbulkload\fR  -m
.DE
.SH
DESCRIPTION
.PP
The user may provide a \fIdmformat \fR input file if a data source is providing
data for the first time. An attempt will be made to add each entry
to the directory. A \fInewdmformat\fR input file will be generally be used
during an incremental update.
.PP
The non-interactive mode is intended for initial processing of the
updates files. Any problems that are encountered that can not be resolved 
automatically may then be dealt with in the interactive mode by specifying
the output file from the non-interactive session as the input file.
.PP
When in interactive mode if the input file does not commence with the
\fBnewdmformat\fR statement, the tool will run an initial check
looking for leading \fInewdmformat\fR characters. If any are found
the user will be prompted to confirm that this is a \fInewdmformat\fR
file after all.
.PP
The interactive mode facilitates error correction using
the editor indicated by the user's \fI$EDITOR\fR environment variable.
During initialisation a check is performed to ensure that the variable is defined.
.PP
When pairing off an entry in a data source with an existing Directory
entry, an exact match of a source purported name and a Directory RDN
is required, else administrative intervention would be necessary.
The exact matching requirement is relaxed when the "-e" option is used.
Once the pairing off has been completed an \fIalgoName\fR attribute will
be added to the directory entry so that when the data source
supplies future updates the pairing off can be achieved automatically.
.PP
The \fBbulkmanagement\fR functionality may be accessed when the "-m"
command line option is used.
.KS
.SH
OPTIONS
.LP
-o [i|n]
.IP
The operational mode required. This may be either \fBI\fRnteractive
or \fBN\fRon-interactive.
If the operational mode parameter is absent the user will be prompted 
for it at the beginning of execution.
.LP
-t tailorfilename
.IP
Allows the specification of an alternative bulkload tailor file. Otherwise
the default \fIbulkloadtailor\fR within \fIetc/bulk\fR will be used.
.LP
-u "user name"
.LP
-p password
.IP
The distinguished name of the user's Directory entry and the corresponding
password. If they are not both supplied then the \fIbulkloadtailor\fR file
is examined for them, before a final check for the two parameters within the
\fI.quipurc\fR file.
.LP
-e
.IP
When applying a source that is to update existing Directory entries,
this option relaxes the level of match required before a source
purported name is paired with an existing Directory entry without the need
for user intervention. As long as only a single existing entry is found with 
a \fIcommonName\fR attribute having the same initials and surname, together
with at least a possible match between any title values, 
the pairing off is carried out automatically.
Note: Once pairing off has occurred
an \fIalgoName\fR attribute will be added to the Directory entry, thus
allowing fully automatic entry matching during subsequent updates.
.IP
As an example if an email data source has a purported name of
"cn=Mr John Paul Smith" then, if only a single existing directory entry
has a common name with the "Smith" surname and the initials "J P",
the entries are automatically paired off.
If the directory entry has title information within a \fIcommonName\fR or a
\fIpersonalTitle\fR attribute it must be
a male title or a non-gender specific title (e.g. "Dr").
.KE
.KS
.LP
-d
.IP
This option disables \fBbulkload's\fR loading and usage of the bulk load
attributes \fIalgoName\fR, \fIdeleteEntryAfter\fR,
\fInonMasterUpdatedAttrType\fR and \fIvalidationCheckAt\fR attributes.
These attributes must still be listed in the oidtables.
Sub-directory search lists may still be used. Given the absence of
\fInonMasterUpdatedAttrType\fR attributes no clearing down of attribute values
will occur when a master source adds values for the first time.
Usage of this option in conjunction with "-m" severely reduces the
management functionality that is provided.
.LP
-m
.IP
This command line option causes \fBbulkload\fR to enter the
\fBbulkmanagement\fR mode of operation. \fBbulkmanagement\fR is described in
a later section.
.KE
.KS
.SH
FILE USAGE
.PP
.B Bulkload
creates three journal files for each input file containing :-
.IP i)
Those updates that have been applied successfully to the directory.
.IP ii)
Those updates that require no further action, e.g. the addition of
an attribute value that already exists.
.IP iii)
Those updates that can not be resolved without (further)
administrative intervention.
.LP
The three files have file extensions ".done", ".noAction" and ".PROBLEM"
respectively.
.KE
.SH
ERROR HANDLING CONFIGURABILITY
.PP
When used in interactive mode, unless specifically disabled within
the \fIbulkloadtailor\fR file, the run-time configuration details will
be presented in an edit for review and possible amendment.
The configuration file statements that are most likely to be configured
on a per-session basis are those that define the routing of attribute update
problems once
.B bulkload
has determined an error.
The ten errors currently supported in this manner are:-
.IP i)
Add value that already exists.
.IP ii)
Add attribute that already exists.
.IP iii)
Add value when not master source.
.IP iv)
Add attribute when not master source.
.IP v)
Delete value when not master source.
.IP vi)
Delete attribute when not master source.
.IP vii)
Delete value that does not exist.
.IP viii)
Delete attribute that does not exist.
.IP ix)
Create entry when not create source.
.IP x)
Delete entry when not delete source.
.LP
As an example, given the possibility that two data sources hold
the same values for certain attribute types e.g. telephone number,
the user may wish that subsequent attempts to add the same value are routed
to the "no action required file", instead of being flagged for user
intervention or being routed to the problems file.
The user controls this by specifying "NOACTION", "PROBLEM" or "PROCESS"
after each error. When running in a non-interactive session
any errors with the "PROCESS" option are routed to the problems file.
.PP
Another configuration statement concerns the handling of entry
additions. If an attempt is made to add an entry that is already held
within the Directory, the user can configure \fBbulkload\fR to attempt
to add the attribute values within the add entry request to the Directory entry, 
without further user intervention.
.SH
EDITING UPON AN ERROR CONDITION
.PP
When in interactive mode, if an error occurs that has not been
configured to be sent to an output file directly, the user will be
offered the offending attribute line or RDN block for correction.
The error information received from the DSA will be listed. You
may find that the information
is not sufficient to resolve the problem, in which case you are
advised to read the tail of the \fIQUIPU\fR DSA's
.I dsap.log
file (or the equivalent for other directory implementations).
In particular this occurs when either a mandatory attribute or
a required object class are missing, which cause object class violation errors.
.PP
When correcting an error using the editor new attribute lines
may be inserted as required. Any details that do not appear in a comment
can be corrected.
.PP
If the error had not been caused
by a parse failure of the \fIdmformat \fR or \fInewdmformat \fR data,
the user can exit from the editor and then request one of the following
options:-
.IP a)
Process the edit updates and apply them to the Directory.
.IP b)
Override a name clash problem when creating an entry with a similar
name to an existing entry.
.IP c)
Override a create (or delete) entry failure that was blocked
due to the data source not being a create (or delete) source.
.IP d)
Convert an add entry request to add value requests (upon an add entry failure.)
This will also cause an algorithmic name attribute to be added to the Directory entry.
.IP e)
Edit/route each attribute line that is affected by an error individually 
(when editing a sequence of attributes.)
.IP f)
Re-edit the current attribute file.
.IP g)
Route the original to the problems file.
.IP h)
Route the edit to the problems file.
.IP i)
Route the original to the no action required file.
.IP j)
Route the edit to the no action required file.
.IP K)
Display the help information.
.IP l)
Change the configuration details for the current session.
.KS
.SH
SAMPLE EDIT BUFFERS
.PP
There now follows three sample edit buffers that were generated \fBbulkload\fR
under various error conditions.
.SH
Error in configuration details
.PP
The error message contained in the following sample indicates that the
"PROCESS" token has been mis-spelled.
The user should correct the token before saving and exiting from the edit.
.DS L
#
# Please review the current configuration settings
#
#
# Note: THE SETTINGS WILL NOT BE STORED TO "bulkloadtailor"
#
#
# When an error occurs the respective update may be routed to one of
#
#       1) the NOACTION required file
#       2) the PROBLEMs file
#       3) for interactive PROCESSing
#
# Note: this applies for both the interactive and the
# non-interactive operational modes
#
.DE
.TS
;
l c l.
Add value when not master source	:	NOACTIONFILE
Add attribute when not master source	:	NOACTIONFILE
Add value that already exists	:	PROCESS
Add attribute that already exists	:	PORCESS
.TE
.DS L
#ERROR
#ERROR *** Invalid processing token in "Addattributethatalreadyexists : PORCESS"
#ERROR
.DE
.TS
;
l c l.
Delete value when not master source	:	NOACTIONFILE
Delete attribute when not master source	:	NOACTIONFILE
Delete value that does not exist	:	PROCESS
Delete attribute that does not exist	:	PROCESS
Delete entry when not delete source	:	PROCESS
.TE
.KE
.RE
.KS
.SH
Error when adding a \fIdmformat\fR \fBentry\fR
.PP
This error condition is due to the absence of the \fIorganizationalPerson\fR
object class. Since the error flagged by the directory merely indicates
a class violation the message must be used in conjunction with the tail
of the \fIQUIPU\fR DSA's \fIdsap.log\fR file (or the equivalent for other
directory implementations) e.g.
.DS
 2/12 11:18:06 ros.quip 00752 (incads  )  attribute 'organizationalUnitName' not
allowed in the specified objectclass 
.DE
.PP
The user should add an "objectClass=organizationalPerson" attribute
line before saving and exiting from the edit.
Then select the default "P" option to Process the newly corrected buffer
and apply the updates to the directory.
.TS
box;
l.
# Add entry failure due to class violation - check your dsap.log file
# *** Update error - Object class violation ***...

rootedAt= c=GB@o=University College London@ou=Computer Science

cn= David Goodman
cn= D Goodman
sn= Goodman
uid= dgoodman
mail= D.Goodman@cs.ucl.ac.uk
mhsORAddresses= /I=D/S=Goodman/OU=cs/O=ucl/PRMD=UK.AC/ADMD=GOLD 400/C=GB/
userClass= csadmin
roomNumber= G24
personalTitle= Mr
telephoneNumber= +44 71-380-7294
ou= Computer Science
objectClass= person
objectClass= newPilotPerson
objectClass= pilotObject
objectClass= quipuObject
objectClass= top
objectClass= mhsUser
objectClass= telBookObject
phoneBookVersion= 2.0
.TE
.KE
.KS
.SH
Error with a \fInewdmformat\fR \fBattribute sequence update\fR
.PP
In this example \fBbulkload\fR has been configured to notify the
user of any attribute value additions that fail due
to the value already being present in the Directory entry.
The error indicates that the value "+44 71-387-7050 x3687" is already
held for the \fItelephoneNumber\fR attribute.
Any one of the following three courses of action could be taken:-
.IP i)
Delete the attribute line ">telephoneNumber = +44 71-387-7050 x3687",
since the value already exists anyway, save the edit and exit.
Then select the default "P" option to Process the newly corrected buffer
and apply the updates to the directory.
.IP ii)
Exit from the edit and select option "C" to change the configuration
routing information for the statement "Add value that already exists",
from "PROCESS" to "NOACTIONFILE" (or "PROBLEM").
Save the configuration edit and then select the default "P" option which
will lead to the attribute line that caused the error being routed to the
".noAction" (or ".PROBLEM") file, before the remaining attribute
updates are re-applied to the Directory.
.IP iii)
Exit from the edit and select option "I" to handle lines in error individually.
An edit will then be displayed containing just the attribute line in error.
If you exit from this edit you can route the attribute line to an output file 
directly.
.sp 2
.RE
.TS
box;
l.
# Attribute update rootedAt : c=GB@o=University College London@ou=Computer Science
# Attribute update for RDN  : cn= Kevin McCarthy
# Please correct the following errors
# ...
+favouriteDrink  = Milk
# **** Attribute error when master source ****
# Type or value already exists...
>telephoneNumber = +44 71-387-7050 x3687
<telephoneNumber = +44 71-387-7050 x9999
.TE
.KE
.bp
.SH
Bulkmanagement functionality
.PP
Six functions are supported:-
.IP 1)
Report on master/create and delete sources.
.IP 2)
Check for expired entries.
.IP 3)
Dump Directory Subtree.
.IP 4)
Reset the validationCheckAt attributes within a Subtree.
.IP 5)
Compare a data source to the Directory.
.IP 6)
Check for entries that have not be compared against.
.SH
1) Report on master/create and delete sources.
.PP
The user will be required to input the parent distinguished name
for a Directory sub-tree. The tool will then search for all the
\fIorganization\fR, \fIorganizationalUnit\fR and \fIlocality\fR entries within
the sub-tree. For each one a list of create entry and delete entry
sources, together with the master sources for each attribute type,
will be output to a report file.
.SH
2) Check for expired entries.
.PP
Within a user specified Directory sub-tree a search for any entries that
have a \fIdeleteEntryAfter\fR value earlier than the user specified expiry
time (the default being the current time) will be instituted. The results
will then be listed in an edit buffer so that any entries that are to
be retained can be removed from the buffer before saving. The user
will then be prompted to confirm that the remaining entries are to
be removed. The attributes held by entries that are to be deleted are
written to the log file, in the \fIdmformat\fR, if deletion is successful.
.SH
3) Dump Directory Subtree.
.PP
List operations are used to retrieve the entries contained within a user
specified sub-tree of the Directory. The results are written in the
\fIdmformat\fR to the report file.
.SH
4) Reset the \fIvalidationCheckAt\fB attributes within a Subtree\fR
.PP
The \fIvalidationCheckAt\fR attributes need to be reset so that
any entries that have not been updated by any of the data sources 
that are applied using function 5 can be determined when using function 6.
.SH
5) Compare a data source to the Directory.
.PP
The complete sets of each of the data sources that have previously 
been used to  update a given sub-directory are read one by one. For each data source 
entry, a search is performed to find the corresponding
algorithmic name attribute within the Directory. If a Directory
entry is found then the source entry's attributes are compared
against it to determine if any are missing.
The Directory entry's \fIvalidationCheckAt\fR attribute is updated to the current time.
.PP
The data from any entries that are held within a data source, but
not within the Directory, together with any missing attributes 
are written to a \fIdmformat\fR report file.
.SH
6) Check for entries that have not be compared against.
.PP
Report on those entries that have not been compared against for any of
the data sources that were used with function 5.
.PP
The UTC time used in the comparison should be the one set using function 4.
If functions 4, 5 and 6 are being run in a single session the UTC time
presented will be that set in function 4.
.bp
.SH
System requirements
.PP
.B Bulkload
requires the presence of the "oidtable" details that are listed in appendix III
in the files "oidtable.oc", oidtable.at" and "oidtable.gen" within the ISODE
etc directory.
The files "bulkloadtailor", "bulkloadwelcome", "bulkloadhelp" and
"bulkmanhelp" need to be held within a "bulk" sub-directory of the ISODE etc directory.
.PP
The editor is defined by the user's \fI$EDITOR\fR environment variable
which should be set to the user's preferred editor.
.SH
Important reminders on planning the bulk load infrastructure
.PP
When planning the \fIdataSource\fR entries required it is essential that the
current settings of the configuration statements listed are reviewed:-
.IP I)
"Highest directory level to search for dataSource entries"
.IP II)
"All sources are masters for"
.IP III)
"When no master specified for a type all sources are masters"
.IP IV)
"All sources are create sources"
.IP V)
"Existing attr value may be cleared down if master source applies 1st value"
.PP
When running the bulkloader with sub-directories that have not been loaded
with \fBbulkload\fR, under a given management infrastructure, 
the affects of the absence of \fInonMasterUpdatedAttrType\fR values must be
considered. Without them all existing attribute will be considered to
have been loaded by a master data source.
.bp
.SH
Appendix II : Bulk load object classes
.b
.LP
bulkDataSourceDetails \fBOBJECT-CLASS\fR
.LP
.B
	SUBCLASS OF
.R
top
.LP
.B
	MUST CONTAIN
.R
{
.LP
			dataSource, deleteSource }
.LP
.B
	MAY CONTAIN
.R
{
.LP
			createSource, masters, nonMasterUpdatedAttrType }
.LP
	::= { bulkObjectClass 1 }
.sp 2
.LP
dataSource \fBATTRIBUTE\fR
.LP
	\fBWITH ATTRIBUTE-SYNTAX\fR
.LP
		caseIgnoreStringSyntax
.LP
::= {bulkAttributeType.5}
.sp
.LP
masters \fBATTRIBUTE\fR
.LP
	\fBWITH ATTRIBUTE-SYNTAX\fR
.LP
		caseIgnoreStringSyntax
.LP
::= {bulkAttributeType.6}
.sp
.LP
deleteSource \fBATTRIBUTE\fB
.LP
	\fBWITH ATTRIBUTE-SYNTAX\fR
.LP
		booleanSyntax
.LP
::= {bulkAttributeType.7}
.sp
.LP
createSource \fBATTRIBUTE\fR
.LP
	\fBWITH ATTRIBUTE-SYNTAX\fR
.LP
		booleanSyntax
.LP
::= {bulkAttributeType.8}
.bp
.LP
bulkDataEntryDetails \fBOBJECT-CLASS\fR
.LP
.B
	SUBCLASS OF
.R
top
.LP
.B
	MAY CONTAIN
.R
{
.LP
			algoName, bulkEntryClass, deleteEntryAfter, 
.LP
			nonMasterUpdatedAttrType, validationCheckAt }
.LP
	::= { bulkObjectClass 2 }
.sp 2
.LP
algoName \fBATTRIBUTE\fR
.LP
	\fBWITH ATTRIBUTE-SYNTAX\fR
.LP
		AlgoNameSyntax
.LP
::= {bulkAttributeType.10}
.sp
.LP
bulkEntryClass \fBATTRIBUTE\fR
.LP
	\fBWITH ATTRIBUTE-SYNTAX\fR
.LP
		caseIgnoreStringSyntax
.LP
::= {bulkAttributeType.20}
.sp
.LP
deleteEntryAfter \fBATTRIBUTE\fR
.LP
	\fBWITH ATTRIBUTE-SYNTAX\fR
.LP
		uTCTimeSyntax
.LP
::= {bulkAttributeType.11}
.sp
.LP
nonMasterUpdatedAttrType \fBATTRIBUTE\fR
.LP
	\fBWITH ATTRIBUTE-SYNTAX\fR
.LP
		caseIgnoreStringSyntax
.LP
::= {bulkAttributeType.12}
.sp
.LP
validationCheckAt \fBATTRIBUTE\fR
.LP
	\fBWITH ATTRIBUTE-SYNTAX\fR
.LP
		uTCTimeSyntax
.LP
::= {bulkAttributeType.13}
.bp
.LP
bulkSubDirListClass \fBOBJECT-CLASS\fR
.LP
.B
	SUBCLASS OF
.R
top
.LP
.B
	MUST CONTAIN
.R
{
.LP
			bulkSubDirList, bulkSubTreeDN, bulkSubTreeDNextension }
.LP
.B
	MAY CONTAIN
.R
{
.LP
			nonMasterUpdatedAttrType }
.LP
	::= { bulkObjectClass 3 }
.sp 2
.LP
bulkSubDirList \fBATTRIBUTE\fR
.LP
	\fBWITH ATTRIBUTE-SYNTAX\fR
.LP
		caseIgnoreStringSyntax
.LP
::= {bulkAttributeType.25}
.sp
.LP
bulkSubTreeDN \fBATTRIBUTE\fR
.LP
	\fBWITH ATTRIBUTE-SYNTAX\fR
.LP
		distinguishedName
.LP
::= {bulkAttributeType.26}
.sp
.LP
bulkSubTreeDNextension \fBATTRIBUTE\fR
.LP
	\fBWITH ATTRIBUTE-SYNTAX\fR
.LP
		distinguishedName
.LP
::= {bulkAttributeType.27}
.bp
.SH
Appendix III : Bulk load oidtable additions
.SH
oidtable.gen
.LP
.TS
;
l l.
bulk:	ucl.106

bulkAttributeType:	bulk.1
bulkAttributeSyntax:	bulk.2
bulkObjectClass:	bulk.3
.TE
.KS
.SH
oidtable.oc
.LP
# Bulk data object classes..
.LP
#dataSourceEntry is the entry RDN..
.LP
bulkDataSourceEntry: bulkObjectClass.0: top: dataSourceEntry :
.DS L
		sourceLoadedAt, lastModifiedTime, lastModifiedBy, acl
.DE
.LP
# dataSource is the entry RDN..
.LP
bulkDataSourceDetails: bulkObjectClass.1: top: dataSource, deleteSource :
.LP
		createSource, masters, nonMasterUpdatedAttrType, lastModifiedTime, lastModifiedBy, acl
.LP
bulkDataEntryDetails:   bulkObjectClass.2: top::
.LP
		algoName, deleteEntryAfter, nonMasterUpdatedAttrType, validationCheckAt
.LP
# bulkSubDirList is the entry RDN..
.LP
bulkSubDirListClass:	bulkObjectClass.3: top: bulkSubDirList, bulkSubTreeDN, 
.LP
		bulkSubTreeDNextension :
.LP
		description, nonMasterUpdatedAttrType, lastModifiedTime, lastModifiedBy, acl
.KE
.KS
.SH
oidtable.at
.LP
# Bulk data attributes..
.TS
;
l l l.
dataSourceEntry:	bulkAttributeType.0	:CaseIgnoreString
sourceLoadedAt:	bulkAttributeType.1	:UTCTime

dataSource:	bulkAttributeType.5	:CaseIgnoreString
masters:	bulkAttributeType.6	:CaseIgnoreString
deleteSource:	bulkAttributeType.7	:Boolean
createSource:	bulkAttributeType.8	:Boolean
algoName:	bulkAttributeType.10	:AlgoNameSyntax

deleteEntryAfter:	bulkAttributeType.11	:UTCTime
masterUpdatedAttrType:	bulkAttributeType.12	:CaseIgnoreString
validationCheckAt:	bulkAttributeType.13	:UTCTime
nonMasterUpdatedAttrType:	bulkAttributeType.14	:CaseIgnoreString

bulkEntryClass:	bulkAttributeType.20	:CaseIgnoreString

bulkSubDirList:	bulkAttributeType.25	:CaseIgnoreString
bulkSubTreeDN:	bulkAttributeType.26	:DN
bulkSubTreeDNextension:	bulkAttributeType.27	:DN
.TE
.SH
Reference
.IP [1]
Paul Barker,
.I
Managing data derived from multiple sources in an X.500 Directory.
.R
.KE
