#include <string.h>
#include <malloc.h>


#include <isode/quipu/name.h>
#include <isode/quipu/ds_search.h>
#include <isode/quipu/dua.h>

#include "util.h"
#include "namelist.h"
#include "filter.h"
#include "readsearch.h"
#include "bulkloadglobal.h"
#include "bulkloadattrnames.h"
#include "configuration.h"
#include "processandupdate.h"
#include "processbulkandupdatedir.h"

extern	struct bulkLoadConfigStruct bulkLoadConfig ;

/*
 * Initialised two lists of attributes types to be searched for and
 * those types to be retrieved..
 */
 
void
initBulkLoadRetrievalAttrs( bulkDetailsToBeFoundpp, bulkDetailsToBeReturnedpp )
struct	namelist ** bulkDetailsToBeFoundpp, ** bulkDetailsToBeReturnedpp ;
{
	addToList( bulkDetailsToBeFoundpp, BL_DATA_SOURCE ) ;

	addToList( bulkDetailsToBeReturnedpp, BL_DATA_SOURCE ) ;
	addToList( bulkDetailsToBeReturnedpp, BL_MASTERS ) ;
	addToList( bulkDetailsToBeReturnedpp, BL_CREATE_SOURCE ) ;
	addToList( bulkDetailsToBeReturnedpp, BL_DELETE_SOURCE ) ;

	return ;
}


void
free_BulkDirectoryDetailspp( baseDataSourceLevelpp )
struct 	dataSourceLevel   ** baseDataSourceLevelpp ;
{
struct	dataSourceLevel	  * currDSLp, * prevDSLp ;
struct 	dataSourceDetails * currDSDp, * prevDSDp ;
struct 	mastersAttrs      * currMAp,  * prevMAp ;


	if( !baseDataSourceLevelpp || !(*baseDataSourceLevelpp) )
	    return ;
	else
	    currDSLp = *baseDataSourceLevelpp ;

	while( currDSLp )
	{
	    currDSDp = currDSLp -> thisLevelsDataSourceDetailsp ;


	    while( currDSDp )
	    {
		if( currDSDp -> dataSourceNamecp != NULL )
	    	    free( currDSDp -> dataSourceNamecp ) ;
	    
	     	currMAp = currDSDp -> mastersAttrsp ;
		
	    	while( currMAp )
	    	{
		    AttrT_free( currMAp -> at ) ;
		
		    prevMAp = currMAp ;

		    currMAp = currMAp -> nextMastersAttrsp ;

		    free( (char *)prevMAp ) ; /* Return node storage */
	    	}

	        prevDSDp = currDSDp ;

	        currDSDp = currDSDp -> nextDataSourceDetailsp ;

	        free( (char *)prevDSDp ) ; /* Return node storage */
	    }

	    prevDSLp = currDSLp ;

	    currDSLp = currDSLp -> nextDataSourceLevelp ;

	    free( (char *)prevDSLp ) ; /* Return node storage */
	}

	*baseDataSourceLevelpp = NULLSOURCELEVEL ;
	
	return ;
}

/*
 * The "masters" attribute value is an attribute type e.g. "telephoneNumber".
 */

createMastersAttrsNode( currMastersAttrspp, mastersAv)
struct	mastersAttrs ** currMastersAttrspp ;
AttributeValue mastersAv ;
{
int	ret_val = CREATE_OK ;
struct	mastersAttrs *  currMastersAttrsp ;
char *	mastersFromAttrVcp ;

	/* Create the node storage */
	if( (currMastersAttrsp = (struct mastersAttrs *)malloc(
					(unsigned)sizeof( struct mastersAttrs ))) != NULLMASTERS)
	{
	    currMastersAttrsp -> nextMastersAttrsp = NULLMASTERS ;

	    *currMastersAttrspp = currMastersAttrsp ;

	    mastersFromAttrVcp = AttrV2pstr( mastersAv ) ;
#ifdef DEBUG
fprintf( stderr, "Masters Av %s\n", mastersFromAttrVcp ) ;
#endif
	    currMastersAttrsp -> at = str2AttrT( mastersFromAttrVcp) ;

	    free( mastersFromAttrVcp ) ;
	}
	else 
	{
	    fprintf( stderr, "Fatal error: memory allocation failed in `createMastersAttrsNode()`\n" );
	    ret_val = CREATE_FATAL ;
	}
	
	return( ret_val ) ;
}


	    
int
createDataSourceDetailsNode( currDataSourceDetailspp, nameListNodep )
struct  dataSourceDetails ** currDataSourceDetailspp ;
struct  namelist * nameListNodep ;
{
int     ret_val = CREATE_OK ;
struct  dataSourceDetails *  currDataSourceDetailsp ;
struct	mastersAttrs      ** currMastersAttrspp ;
Attr_Sequence  as  ;
AV_Sequence    avs ;
RDN	lastRDN = NULLRDN ;
static	AttributeType	dataSourceat = NULLAttrT, TrueAt = NULLAttrT ;
static	AttributeValue  TrueAv = NULLAttrV ;

	if( dataSourceat == NULLAttrT )
	    dataSourceat  = str2AttrT( BL_DATA_SOURCE) ;

	if( TrueAt == NULLAttrT )
	{
	    if( (TrueAt = str2AttrT( BL_DELETE_SOURCE )) != NULLAttrT )
	    {
		TrueAv  = str2AttrV( "TRUE", TrueAt -> oa_syntax ) ;	    
	    }
	}

        /* Create the node storage */
	if( (currDataSourceDetailsp = (struct dataSourceDetails*)malloc(
		                       	(unsigned)sizeof( struct dataSourceDetails ))) != NULLSOURCEDETAILS )
	{
	    if( TrueAt != NULLAttrT && TrueAv != NULLAttrV )
	    {
	    	currDataSourceDetailsp -> nextDataSourceDetailsp = NULLSOURCEDETAILS ;
	    
	    	*currDataSourceDetailspp = currDataSourceDetailsp ;

	    	currMastersAttrspp = &( currDataSourceDetailsp -> mastersAttrsp ) ;
	    
		*currMastersAttrspp = NULLMASTERS ;

		currDataSourceDetailsp -> createSource = TRUE ; /* Default for this optional attribute */

		/* Check if the "dataSource" attribute is within the RDN or
		   if it is listed in the attribute list                    */
		   
		if( (lastRDN = findLastRDN( nameListNodep -> entryDN )) != NULLRDN &&
		     dataSourceat != NULLAttrT &&
		     findAttrTypeRdnComp( dataSourceat, lastRDN )       != NULLRDN   )
		{
		    currDataSourceDetailsp -> dataSourceNamecp  = strdup( nameListNodep -> name ) ;
		}
		else currDataSourceDetailsp -> dataSourceNamecp = NULLCP ;

#ifdef DEBUG
fprintf( stderr, "Source Name %s\n", currDataSourceDetailsp -> dataSourceNamecp ) ;
#endif
	    	for( as = nameListNodep->ats ; ret_val == CREATE_OK &&
				       	       as != NULLATTR ; as = as -> attr_link )
	    	{
		    for( avs = as->attr_value ; ret_val == CREATE_OK &&
		    			        avs != NULLAV ; avs = avs -> avseq_next )
		    {
			if( currDataSourceDetailsp -> dataSourceNamecp == NULLCP &&
			    strcmp( attr2name( as->attr_type, OIDPART), BL_DATA_SOURCE ) == 0 )
			{
			    currDataSourceDetailsp -> dataSourceNamecp = AttrV2pstr( &(avs->avseq_av) ) ;
			}
		    	else if( strcmp( attr2name( as->attr_type, OIDPART), BL_CREATE_SOURCE ) == 0 )
		        {
			    currDataSourceDetailsp -> createSource = 
			        ( AttrV_cmp( TrueAv, &(avs->avseq_av)) == 0 ) ? TRUE : FALSE ;
			}
		    	else if( strcmp( attr2name( as->attr_type, OIDPART), BL_DELETE_SOURCE ) == 0 )
		        {
			    currDataSourceDetailsp -> deleteSource = 
			        ( AttrV_cmp( TrueAv, &(avs->avseq_av)) == 0 ) ? TRUE : FALSE ;
			}
		    	else if( strcmp( attr2name( as->attr_type, OIDPART), BL_MASTERS ) == 0 )
		    	{
			    ret_val = createMastersAttrsNode( currMastersAttrspp, &(avs->avseq_av)) ;

		    	    currMastersAttrspp = &((*currMastersAttrspp) -> nextMastersAttrsp ) ;
			}
		    }
	    	}
	    }
	    else
	    {
	    	fprintf( stderr, "Fatal error: couldnot create a `boolean' attribute\nin `createDataSourceDetailsNode()'\nSuspect oidtable problem\n" );
		ret_val = CREATE_FATAL ;
	    }
	}
	else 
	{
	    fprintf( stderr, "Fatal error: memory allocation failed in `createDataSourceDetailsNode()`\n" );
	    ret_val = CREATE_FATAL ;
	}
	
	return( ret_val ) ;
}



int
createDataSourceLevelNode( currDataSourceLevelpp, searchListp )
struct 	dataSourceLevel ** currDataSourceLevelpp ;
struct 	namelist * searchListp ;
{
int	ret_val = CREATE_OK ;
struct 	dataSourceLevel   *  currDataSourceLevelp ;
struct	dataSourceDetails ** currDataSourceDetailspp ;
struct  namelist * nameListNodep ;

	/* Create the node storage */
	if( (currDataSourceLevelp = (struct dataSourceLevel *)malloc(
						(unsigned)sizeof( struct dataSourceLevel ))) != NULLSOURCELEVEL)
	{
	    currDataSourceLevelp -> nextDataSourceLevelp = NULLSOURCELEVEL ;
	    
	    *currDataSourceLevelpp  = currDataSourceLevelp ;

	    currDataSourceDetailspp = &(currDataSourceLevelp->thisLevelsDataSourceDetailsp) ;

	    *currDataSourceDetailspp = NULLSOURCEDETAILS ;

	     /* Extract the attributes from the search list.. */

	    for( nameListNodep = searchListp ; ret_val == CREATE_OK &&
	    				       nameListNodep!=NULLLIST ; nameListNodep = nameListNodep -> next )
	    {
	    	ret_val = createDataSourceDetailsNode( currDataSourceDetailspp, nameListNodep ) ;

	    	currDataSourceDetailspp = &((*currDataSourceDetailspp)->nextDataSourceDetailsp) ;
	    }
	}
	else 
	{
	    fprintf( stderr, "Fatal error: memory allocation failed in `createDataSourceLevelNode()`\n" );
	    ret_val = CREATE_FATAL ;
	}
	
	return( ret_val ) ;
}


int
defineAndActionBulkDetailsSearch( parentDN, attrsToBeFoundp, attrsToBeReturnedp, searchListpp )
DN	parentDN ;
struct 	namelist  * attrsToBeFoundp, * attrsToBeReturnedp ;
struct 	namelist ** searchListpp ;
{
int	ret_val = SEARCHOK ;
struct	ds_search_arg sarg;
struct	ds_search_result sresult;
struct	DSError serror ;


	sarg = * fillMostSearchArgs( parentDN, SRA_ONELEVEL, attrsToBeReturnedp, FALSE ) ;

	makePresentFilter( &sarg.sra_filter, attrsToBeFoundp ) ;


	/* Carry out the search and produce a list of results.. */
	ret_val = RetrieveAndListAttrs( searchListpp, &sarg, &sresult, &serror ) ;


        if (ret_val != SEARCHOK)
	{
	    logListSuccess( LIST_ERROR, "bulkDataDetails", 0) ;
	    ds_error_free( &serror ) ;
	}
	else
	    logListSuccess( LIST_OK, "bulkDataDetails", listlen( *searchListpp )) ;


	freeSearchArgs( &sarg ) ;

	return( ret_val ) ; /* SEARCHNOTOK, SEARCHOK,... [bulkloadglobal.h] */
}



/*
 * Generate a search requests to retrieve the bulk data details
 * contained within the directory.
 *
 * Note: the procedure unlinks an increasing number of the lowest
 * "dncomp" nodes so that the "parentDN" points to progressively
 * higher directory levels.
 */

int
retrieveBulkDirectoryDetails( parentDN, baseDataSourceLevelpp )
DN	parentDN ;
struct 	dataSourceLevel  ** baseDataSourceLevelpp ; /* Return parameter */
{
struct 	dataSourceLevel  ** currDataSourceLevelpp = baseDataSourceLevelpp ;
struct	namelist *searchListp ;
int	ret_val = OK, res ;
int	compCount, postUnlinkDNcomp, firstLoopIteration=TRUE ;
DN	DNcompp = NULLDN , bufferedDNcompp = NULLDN ;

static	struct	namelist  * bulkDetailsToBeFoundp    = NULLLIST ;
static	struct	namelist  * bulkDetailsToBeReturnedp = NULLLIST ;
static	int 	attrsInitialised = FALSE ;

#ifdef DEBUG
char * 	cp ;
#endif

	if( attrsInitialised == FALSE )
	{
	    initBulkLoadRetrievalAttrs( &bulkDetailsToBeFoundp, &bulkDetailsToBeReturnedp ) ;

	    attrsInitialised =  TRUE ;
	}

	
	postUnlinkDNcomp = countDNcomps( parentDN ) ;

	/* Retrieve the "dataSource" entries from the directory 
	   level by level, commencing at the rootedAt level..   */

	while( parentDN != NULLDN && ret_val == OK &&
			postUnlinkDNcomp >= bulkLoadConfig.highestDataSourceSearchLevel )
	{
	    /* If this is not the first loop iteration */
	    if( firstLoopIteration == FALSE )
	    {
		/* Put back previously unlinked "dncomps".. */

		if( bufferedDNcompp != NULLDN )
		    DNcompp -> dn_parent = bufferedDNcompp ; /* relink trailing dncomp(s) */
		    
		/* Unlink the required number of "dncomp" nodes.. */

		DNcompp = parentDN ; /* .. leading dncomp of "parentDN" */
		
		for( compCount = 1 ; compCount < postUnlinkDNcomp ; compCount++ )
		{
		    DNcompp = DNcompp -> dn_parent ;
		}

		/* DNcompp now point to the "postUnlink"'th node */

		bufferedDNcompp = DNcompp -> dn_parent ;

		DNcompp -> dn_parent = NULLDN ; /* Unlink the trailing dncomp(s) */
	    }

#ifdef DEBUG
cp = dn2pstr( parentDN ) ;
if( cp != NULLCP )
{
fprintf( stderr, "parent dn cp %s\n", cp ) ;
free(cp) ;
}
#endif

#ifdef DEBUG
fprintf( stderr, "Comp num = %d\n", postUnlinkDNcomp ) ;
#endif
	    
	    if( defineAndActionBulkDetailsSearch( parentDN, bulkDetailsToBeFoundp,
				 bulkDetailsToBeReturnedp, &searchListp ) == SEARCHOK )
	    {
		/* Confirm that results have been found.. */
		if( searchListp != NULLLIST )
		{
	    	    res = createDataSourceLevelNode( currDataSourceLevelpp, searchListp ) ;

		    if( res == CREATE_OK )
		    {
		    	currDataSourceLevelpp = &((*currDataSourceLevelpp) -> nextDataSourceLevelp) ;
	     	    }
		    else ret_val = NOTOK ;

	     	    freeListpp( &searchListp ) ;
		}

		--postUnlinkDNcomp ;
	    }
	    else ret_val = NOTOK ;

	    firstLoopIteration = FALSE ;

	} /* while loop */

	/* Put back previously unlinked "dncomps".. */

	if( bufferedDNcompp != NULLDN )
	    DNcompp -> dn_parent = bufferedDNcompp ; /* relink trailing dncomp(s) */
		    
	return( ret_val ) ; /* OK, NOTOK */
}

/*
 * Search through the bulk details linked list to find the lowest
 * level create source(s) and check if the specified source is one of them.
 *
 * If there is a dataSource attribute within the RDN all sources are create sources for 
 * the corresponding entry.
 */

int
sourceIsACreateSource( baseDataSourceLevelp, sourceNamecp, entryRDN )
struct 	dataSourceLevel   * baseDataSourceLevelp ;
char *	sourceNamecp ;
RDN	entryRDN ;
{
int	isCreateSource = FALSE, createSourceFound = FALSE ;
struct 	dataSourceLevel   * currDataSourceLevelp = baseDataSourceLevelp ;
struct  dataSourceDetails * currDataSourceDetailsp ;
char *	sourcecp ;
static	AttributeType	dataSourceat = NULLAttrT, bulkSubDirListat = NULLAttrT ;

	if( bulkLoadConfig.disableBulkAttributeUsage  == TRUE ||
	    bulkLoadConfig.allSourcesAreCreateSources == TRUE   )
	    return( TRUE ) ;
	
	if( dataSourceat == NULLAttrT )
	    dataSourceat =  str2AttrT( BL_DATA_SOURCE ) ;

	if( bulkSubDirListat == NULLAttrT )
	    bulkSubDirListat =  str2AttrT( BL_BULK_SUB_DIR_LIST ) ;

	if( entryRDN != NULLRDN && (findAttrTypeRdnComp( dataSourceat    , entryRDN ) != NULLRDN ||
				    findAttrTypeRdnComp( bulkSubDirListat, entryRDN ) != NULLRDN   ))
	    return( TRUE ) ;


	while( currDataSourceLevelp != NULLSOURCELEVEL && createSourceFound == FALSE )
	{
	    currDataSourceDetailsp = currDataSourceLevelp -> thisLevelsDataSourceDetailsp ;
	
	    while( currDataSourceDetailsp != NULLSOURCEDETAILS && isCreateSource == FALSE )
	    {
		if( currDataSourceDetailsp -> createSource == TRUE )
		{
		    createSourceFound = TRUE ;

		    /* Extract the source name component from its DN string.. */
		    sourcecp = strrchr( currDataSourceDetailsp -> dataSourceNamecp, '=' ) ;

		    if( sourcecp != NULLCP )
		    {
		    	++sourcecp ; /* Skip the '=' to get to the last component */

			/* Skip any leading {T.61} component.. */
			if( strlen( sourcecp ) >= 6 &&
			    *sourcecp == '{' && *(sourcecp + 1) == 'T' && *(sourcecp + 2) == '.' &&
			    *(sourcecp + 3) == '6' && *(sourcecp + 4) == '1' && *(sourcecp + 5) == '}' )
			{
			    sourcecp += 6 ;
			}

			if( strcasecmp( sourceNamecp, sourcecp ) == 0 )
		    	    isCreateSource = TRUE ;
		    }
    		}

		currDataSourceDetailsp = currDataSourceDetailsp -> nextDataSourceDetailsp ;
	    }

	    currDataSourceLevelp = currDataSourceLevelp -> nextDataSourceLevelp ;
	}

	return( isCreateSource ) ;
}

void
reportCreateSource( sourceReportfp, baseDataSourceLevelp )
FILE *	sourceReportfp ;
struct 	dataSourceLevel * baseDataSourceLevelp ;
{
int	createSourceFound = FALSE ;
struct 	dataSourceLevel   * currDataSourceLevelp = baseDataSourceLevelp ;
struct  dataSourceDetails * currDataSourceDetailsp ;
char *	sourcecp ;

	fprintf( sourceReportfp, "Create sources :-\n" ) ;

	if( bulkLoadConfig.disableBulkAttributeUsage  == TRUE ||
	    bulkLoadConfig.allSourcesAreCreateSources == TRUE   )
	{
	    fprintf( sourceReportfp, "*** ALL SOURCES ARE CREATE SOURCES ***\n\n" ) ;
	    fprintf( sourceReportfp, "--------------------------------------\n" ) ;
	    return ;
	}

	fprintf( sourceReportfp, "All sources are create sources for dataSource and bulkSubDirList entries\n" ) ;

	while( currDataSourceLevelp != NULLSOURCELEVEL && createSourceFound == FALSE )
	{
	    currDataSourceDetailsp = currDataSourceLevelp -> thisLevelsDataSourceDetailsp ;
	
	    while( currDataSourceDetailsp != NULLSOURCEDETAILS )
	    {
		if( currDataSourceDetailsp -> createSource == TRUE )
		{
		    createSourceFound = TRUE ;

		    /* Extract the source name component from its DN string.. */
		    sourcecp = strrchr( currDataSourceDetailsp -> dataSourceNamecp, '=' ) ;

		    if( sourcecp != NULLCP )
		    {
			fprintf( sourceReportfp, "\n%s\n", currDataSourceDetailsp -> dataSourceNamecp ) ; 
		    }
    		}

		currDataSourceDetailsp = currDataSourceDetailsp -> nextDataSourceDetailsp ;
	    }

	    currDataSourceLevelp = currDataSourceLevelp -> nextDataSourceLevelp ;
	}

	fprintf( sourceReportfp, "------------------------------------\n" ) ;

	return ;
}


/*
 * Search through the bulk details linked list to find the lowest level
 * delete source(s) and check if the specified source is one of them..
 */

int
sourceIsADeleteSource( baseDataSourceLevelp, sourceNamecp )
struct 	dataSourceLevel   * baseDataSourceLevelp ;
char *	sourceNamecp ;
{
int	isDeleteSource = FALSE, deleteSourceFound = FALSE ;
struct 	dataSourceLevel   * currDataSourceLevelp = baseDataSourceLevelp ;
struct  dataSourceDetails * currDataSourceDetailsp ;
char *	sourcecp ;

	if( bulkLoadConfig.disableBulkAttributeUsage == TRUE )
	    return( TRUE ) ;

	while( currDataSourceLevelp != NULLSOURCELEVEL && deleteSourceFound == FALSE )
	{
	    currDataSourceDetailsp = currDataSourceLevelp -> thisLevelsDataSourceDetailsp ;
	
	    while( currDataSourceDetailsp != NULLSOURCEDETAILS && isDeleteSource == FALSE )
	    {
		if( currDataSourceDetailsp -> deleteSource == TRUE )
		{
		    deleteSourceFound = TRUE ;

		    /* Extract the source name component from its DN string.. */
		    sourcecp = strrchr( currDataSourceDetailsp -> dataSourceNamecp, '=' ) ;

		    if( sourcecp != NULLCP )
		    {
		    	++sourcecp ; /* Skip the '=' to get to the last component */

			/* Skip any leading {T.61} component.. */
			if( strlen( sourcecp ) >= 6 &&
			    *sourcecp == '{' && *(sourcecp + 1) == 'T' && *(sourcecp + 2) == '.' &&
			    *(sourcecp + 3) == '6' && *(sourcecp + 4) == '1' && *(sourcecp + 5) == '}' )
			{
			    sourcecp += 6 ;
			}

			if( strcasecmp( sourceNamecp, sourcecp ) == 0 )
		    	    isDeleteSource = TRUE ;
		    }
    		}

		currDataSourceDetailsp = currDataSourceDetailsp -> nextDataSourceDetailsp ;
	    }

	    currDataSourceLevelp = currDataSourceLevelp -> nextDataSourceLevelp ;
	}

	return( isDeleteSource ) ;
}

void
reportDeleteSource( sourceReportfp, baseDataSourceLevelp )
FILE *	sourceReportfp ;
struct 	dataSourceLevel * baseDataSourceLevelp ;
{
int	deleteSourceFound = FALSE ;
struct 	dataSourceLevel   * currDataSourceLevelp = baseDataSourceLevelp ;
struct  dataSourceDetails * currDataSourceDetailsp ;
char *	sourcecp ;

	fprintf( sourceReportfp, "Delete sources :-\n" ) ;

	if( bulkLoadConfig.disableBulkAttributeUsage == TRUE )
	{
	    fprintf( sourceReportfp, "*** ALL SOURCES ARE DELETE SOURCES ***\n\n" ) ;
	    fprintf( sourceReportfp, "--------------------------------------\n" ) ;
	    return ;
	}

	while( currDataSourceLevelp != NULLSOURCELEVEL && deleteSourceFound == FALSE )
	{
	    currDataSourceDetailsp = currDataSourceLevelp -> thisLevelsDataSourceDetailsp ;
	
	    while( currDataSourceDetailsp != NULLSOURCEDETAILS )
	    {
		if( currDataSourceDetailsp -> deleteSource == TRUE )
		{
		    deleteSourceFound = TRUE ;

		    /* Extract the source name component from its DN string.. */
		    sourcecp = strrchr( currDataSourceDetailsp -> dataSourceNamecp, '=' ) ;

		    if( sourcecp != NULLCP )
		    {
			fprintf( sourceReportfp, "\n%s\n", currDataSourceDetailsp -> dataSourceNamecp ) ; 
		    }
    		}

		currDataSourceDetailsp = currDataSourceDetailsp -> nextDataSourceDetailsp ;
	    }

	    currDataSourceLevelp = currDataSourceLevelp -> nextDataSourceLevelp ;
	}

	if( deleteSourceFound == FALSE )
	   fprintf( sourceReportfp, "*** NO DELETE SOURCES ***\n" ) ;

	fprintf( sourceReportfp, "------------------------------------\n" ) ;

	return ;
}
/*
 * Search through the bulk details linked list to find the lowest
 * level master source for the specified attribute type. At this level
 * check if the specified source is  one of the master sources listed.
 *
 * Note: if no master source is found then the configuration information
 * is examined to determine whether all sources should be treated as a
 * master source.
 */

int
sourceIsMasterOfAttrT( baseDataSourceLevelp, sourceNamecp, at )
struct 		dataSourceLevel * baseDataSourceLevelp ;
char *		sourceNamecp ;
AttributeType	at ;
{
int	isMasterSource = FALSE, aMasterSourceFound = FALSE, attrTindex ;
struct 	dataSourceLevel   * currDataSourceLevelp = baseDataSourceLevelp ;
struct  dataSourceDetails * currDataSourceDetailsp ;
struct  mastersAttrs      * currMastersAttrsp ;
char *	sourcecp ;

#ifdef DEBUG
char *	cp1, * cp2 ;
#endif
	if( bulkLoadConfig.disableBulkAttributeUsage == TRUE )
	    return( TRUE ) ;

	/* Check the list of attribute types that are mastered by all sources first */

	for( attrTindex = 0 ; attrTindex <= bulkLoadConfig.lastUsedAttrTIndex ; attrTindex++ )
	{
	    if( bulkLoadConfig.allSourcesAreMastersForAttrT[ attrTindex ] != NULLAttrT &&
	    	AttrT_cmp( bulkLoadConfig.allSourcesAreMastersForAttrT[ attrTindex ], at ) == 0 )
	    {
	    	isMasterSource = TRUE ; aMasterSourceFound = TRUE ;
		break ;
	    }
	}

	while( currDataSourceLevelp != NULLSOURCELEVEL && aMasterSourceFound == FALSE )
	{
	    currDataSourceDetailsp = currDataSourceLevelp -> thisLevelsDataSourceDetailsp ;

	    while( currDataSourceDetailsp != NULLSOURCEDETAILS && isMasterSource == FALSE )
	    {
	    	currMastersAttrsp = currDataSourceDetailsp -> mastersAttrsp;

		while( currMastersAttrsp != NULLMASTERS && isMasterSource == FALSE )
		{
#ifdef DEBUG
cp1 = AttrT2pstr( at ), cp2 = AttrT2pstr(currMastersAttrsp -> at) ;
fprintf( stderr, "Type %s %s \n", cp1, cp2 ) ;
free( cp1 ), free( cp2 ) ;
#endif
		    /* See if the current masters attribute type is the one required.. */
		    if( AttrT_cmp( at, currMastersAttrsp -> at ) == 0 )
		    {
		    	aMasterSourceFound = TRUE ;
		    	
			/* Extract the source name component from its DN string.. */
		    	sourcecp = strrchr( currDataSourceDetailsp -> dataSourceNamecp, '=' ) ;

		    	if( sourcecp != NULL )
		    	{
		    	    ++sourcecp ; /* Skip the '=' to get to the last component */

			    /* Skip any leading {T.61} component.. */
			    if( strlen( sourcecp ) >= 6 &&
			    	*sourcecp == '{' && *(sourcecp + 1) == 'T' && *(sourcecp + 2) == '.' &&
			    	*(sourcecp + 3) == '6' && *(sourcecp + 4) == '1' && *(sourcecp + 5) == '}' )
			    {
				sourcecp += 6 ;
			    }

		    	    if( strcasecmp( sourceNamecp, sourcecp ) == 0 )
			    	isMasterSource = TRUE ;
			}
		    }

		    currMastersAttrsp = currMastersAttrsp -> nextMastersAttrsp ;
		}

		currDataSourceDetailsp = currDataSourceDetailsp -> nextDataSourceDetailsp ;
	    }

	    currDataSourceLevelp = currDataSourceLevelp -> nextDataSourceLevelp ;
	}

	/* Possible over rule if the attribute type has no master source
	   within the current scope.. */
	   
	if( baseDataSourceLevelp != NULLSOURCELEVEL &&
	    isMasterSource == FALSE && aMasterSourceFound == FALSE &&
	    bulkLoadConfig.noMasterSpecifiedSoAllSourcesAreMasters == TRUE )
	{
	    isMasterSource = TRUE ;
	}

#ifdef DEBUG
cp1 = AttrT2pstr( at ) ;
fprintf( stderr, "MASTER ATTR Type %s \n", cp1 ) ;
free( cp1 ) ;
#endif
	return( isMasterSource ) ;
}

void
addAttrTtoMasterList( AtListp, maxSizeOfAtListp, currTopIndexOfAtListp, at )
AttributeType	** AtListp ;
int  *	maxSizeOfAtListp, * currTopIndexOfAtListp ;
AttributeType	at ;
{
	if( AtListp && *AtListp &&
	    *currTopIndexOfAtListp + 1 == *maxSizeOfAtListp)
	{
	    /* Increase the list size */
	    *maxSizeOfAtListp += 25 ;
	    
	    *AtListp = (AttributeType *)realloc( (AttributeType *)(*AtListp), 
	    					sizeof( AttributeType *) * (*maxSizeOfAtListp)) ;
	}

	if( AtListp && *AtListp && at != NULLAttrT )
	{
	    ++(*currTopIndexOfAtListp) ;

	    *((*AtListp) + (*currTopIndexOfAtListp)) = AttrT_cpy( at ) ;
	}
	
	return ;
}


void
reportMasterSource( sourceReportfp, baseDataSourceLevelp )
FILE *	sourceReportfp ;
struct 	dataSourceLevel * baseDataSourceLevelp ;
{
int	attrTindex, typeFound ;
struct 	dataSourceLevel   * currDataSourceLevelp = baseDataSourceLevelp ;
struct  dataSourceDetails * currDataSourceDetailsp ;
struct  mastersAttrs      * currMastersAttrsp ;
char *	sourcecp, * attrTcp ;
AttributeType *	masteredAtList, * masteredAtThisLevelList, at ;
int	maxSizeOfMasterAtList 	       = 50, currTopIndexOfMasterAtList = -1 ;
int	maxSizeOfMasterAtThisLevelList = 50, currTopIndexOfMasterAtThisLevelList = -1 ;

	fprintf( sourceReportfp, "Master sources :-\n" ) ;

	if( bulkLoadConfig.disableBulkAttributeUsage == TRUE )
	{
	    fprintf( sourceReportfp, "*** ALL SOURCES ARE MASTER SOURCES OF ALL ATTRIBUTES ***\n\n" ) ;
	    fprintf( sourceReportfp, "--------------------------------------------------------\n" ) ;
	    return ;
	}

	/* Initialise the list.. */
	if( (masteredAtList = (AttributeType *)malloc( sizeof( AttributeType ) * maxSizeOfMasterAtList )) == ((AttributeType *)NULL) )
	{
	    fprintf( sourceReportfp, "#ERROR could not malloc the required storage for the mastered attributes\n" );
	    return ;
	}
	
	/* Check the list of attribute types that are mastered by all sources first */

	for( attrTindex = 0 ; attrTindex <= bulkLoadConfig.lastUsedAttrTIndex ; attrTindex++ )
	{
	    if( bulkLoadConfig.allSourcesAreMastersForAttrT[ attrTindex ] != NULLAttrT &&
	    	(attrTcp = AttrT2pstr( bulkLoadConfig.allSourcesAreMastersForAttrT[ attrTindex ] )) != NULLCP )
	    {
	    	addAttrTtoMasterList( &masteredAtList,
 			&maxSizeOfMasterAtList, &currTopIndexOfMasterAtList,
			bulkLoadConfig.allSourcesAreMastersForAttrT[ attrTindex ] ) ;

		fprintf( sourceReportfp, "ALL sources master %s\n", attrTcp ) ;

	    	free( attrTcp ) ;
	    }
	}

	fprintf( sourceReportfp, "\n" ) ;

	/* Initialise the list.. */
	if( (masteredAtThisLevelList = (AttributeType *)malloc( sizeof( AttributeType * ) * maxSizeOfMasterAtThisLevelList )) == ((AttributeType *)NULL) )
	{
	    fprintf( sourceReportfp, "#ERROR could not malloc the required storage for the mastered attributes\n" );
	    return ;
	}

	while( currDataSourceLevelp != NULLSOURCELEVEL )
	{
	    currDataSourceDetailsp = currDataSourceLevelp -> thisLevelsDataSourceDetailsp ;

	    while( currDataSourceDetailsp != NULLSOURCEDETAILS )
	    {
	    	currMastersAttrsp = currDataSourceDetailsp -> mastersAttrsp;

		sourcecp = strrchr( currDataSourceDetailsp -> dataSourceNamecp, '=' ) ;

		if( sourcecp != NULLCP )
		{
		    fprintf( sourceReportfp, "\n%s\n", currDataSourceDetailsp -> dataSourceNamecp ) ;
		}

		while( currMastersAttrsp != NULLMASTERS )
		{
		    if( (at = currMastersAttrsp-> at ) != NULLAttrT )
		    {
		    	typeFound = FALSE ;
		    
			/* Check if the type is already mastered at another level .. */
			for( attrTindex = 0 ; attrTindex <= currTopIndexOfMasterAtList ; attrTindex++ )
		    	{
		    	    if( AttrT_cmp( *(masteredAtList + attrTindex), at ) == 0 )
			    {
			    	typeFound = TRUE ; 
				break ;
			    }		    	
		        }

			if( typeFound == FALSE )
		    	{
			    if( (attrTcp = AttrT2pstr( at )) != NULLCP )
			    {
			    	fprintf( sourceReportfp, "masters %s\n", attrTcp ) ;
				free( attrTcp ) ;
			    }
			    
		    	    /* Check that the attribute type is not already listed at this level.. */
		    	    for( attrTindex = 0 ; attrTindex <= currTopIndexOfMasterAtThisLevelList ; attrTindex++ )
		    	    {
		    	    	if( AttrT_cmp( *(masteredAtThisLevelList + attrTindex), at ) == 0 )
			    	{
			    	    typeFound = TRUE ;
				    break ;
			    	}
			    }

			    if( typeFound == FALSE )
				addAttrTtoMasterList( &masteredAtThisLevelList,
 					&maxSizeOfMasterAtThisLevelList, &currTopIndexOfMasterAtThisLevelList, at ) ;
			}
		    }

		    currMastersAttrsp = currMastersAttrsp -> nextMastersAttrsp ;
		}

		currDataSourceDetailsp = currDataSourceDetailsp -> nextDataSourceDetailsp ;
	    }

	    /* Transfer any master types to the full list.. */
	    for( attrTindex = 0 ; attrTindex <= currTopIndexOfMasterAtThisLevelList ; attrTindex++ )
	    {
	    	if( *(masteredAtList + attrTindex) != NULLAttrT )
		{
		    addAttrTtoMasterList( &masteredAtList,
 				&maxSizeOfMasterAtList, &currTopIndexOfMasterAtList, *(masteredAtThisLevelList + attrTindex) ) ;

		    AttrT_free( *(masteredAtThisLevelList + attrTindex) ) ;
		}
	    }

	    currTopIndexOfMasterAtThisLevelList = -1 ;

	    currDataSourceLevelp = currDataSourceLevelp -> nextDataSourceLevelp ;
	}


	/* Free the attribute type list.. */
	for( attrTindex = 0 ; attrTindex <= currTopIndexOfMasterAtList ; attrTindex++ )
	{
	    if( *(masteredAtList + attrTindex) != NULLAttrT )
		AttrT_free( *(masteredAtList + attrTindex) ) ;
	}

	if( masteredAtThisLevelList )
	    free( (char *)masteredAtThisLevelList ) ;

	if( masteredAtList )
	    free( (char *)masteredAtList ) ;
	
	
	/* Possible over rule if the attribute type has no master source
	   within the current scope.. */
	   
	if( bulkLoadConfig.noMasterSpecifiedSoAllSourcesAreMasters == TRUE )
	    fprintf( sourceReportfp, "\n\n*** ALL OTHER ATTRIBUTE TYPES ARE MASTERED BY ALL SOURCES ***\n\n" ) ;
	
	return ;
}

void
reportSourceDetails( sourceReportfp, baseDataSourceLevelp )
FILE *	sourceReportfp ;
struct 	dataSourceLevel * baseDataSourceLevelp ;
{
	reportCreateSource( sourceReportfp, baseDataSourceLevelp ) ;
	reportDeleteSource( sourceReportfp, baseDataSourceLevelp ) ;
	reportMasterSource( sourceReportfp, baseDataSourceLevelp ) ;
	
	return ;
}
